/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "CTimeShiftManager.h"
#include "stack_aids.h"
#include <assert.h>

CTimeShiftManager::CTimeShiftManager(void)
{
	mAdjustingList 		= false;
	mWarned 			= false;
	mRunning 			= false;
	mNextEventOfInterest= -1;
}


ostream& operator<<(ostream &o,CTimeShiftManager &m)
{
	stack_MutexLock sml(&m.mMutex);
	o << ((int)m.mEventList.size()) << endl;
	for (int x = 0; x < m.mEventList.size(); ++x)
		o << m.mEventList[x] << endl;
	return o;
}


istream& operator>>(istream &i,CTimeShiftManager &m)
{
	m.BeginListAdjustment();
	{
	stack_MutexLock sml(&m.mMutex);
	m.mEventList.erase(m.mEventList.begin(),m.mEventList.end());
	int s;
	i >> s;
	for (int x = 0; x < s; ++x)
		{
		CTimeShiftEvent e;
		i >> e;
		m.mEventList.push_back(e);
		}
	}
	m.EndListAdjustment();
	return i;
}


void
CTimeShiftManager::Reset(void)
{
	BeginListAdjustment();
	EndListAdjustment();
}

void		
CTimeShiftManager::BeginListAdjustment(void)
{
	stack_MutexLock sml(&mMutex);
	assert(!mAdjustingList);

	mWarned 			= false;
	mRunning 			= false;
	mNextEventOfInterest= -1;
	mAdjustingList 		= true;
}

void
CTimeShiftManager::FigureNextEventOfInterest(COleDateTime *current_time)
{
	COleDateTime	ct;

	if (!current_time)

		ct = COleDateTime::GetCurrentTime();

	else

		ct = *current_time;

	// lop off seconds

	ct.SetDateTime(ct.GetYear(),ct.GetMonth(),ct.GetDay(),

		ct.GetHour(),ct.GetMinute(),0);


	mNextEventOfInterest = -1;

	for (int x = 0; x < mEventList.size(); ++x)
		{
		COleDateTime n = mEventList[x].NextActivationTime();

		// lop off seconds

		n.SetDateTime(n.GetYear(),n.GetMonth(),n.GetDay(),

			n.GetHour(),n.GetMinute(),0);


		COleDateTimeSpan s = mEventList[x].Duration();
		if (n >= ct)
			{
			if (mNextEventOfInterest == -1)
				{
				mNextEventOfInterest = x;
				mDurationOfInterest = s;
				mStartTimeOfInterest = n;
				}
			else if (n < mStartTimeOfInterest)
				{
				mStartTimeOfInterest = n;
				mNextEventOfInterest = x;
				mDurationOfInterest = s;
				}
			}
		}
}

void

CTimeShiftManager::RemoveOldEvents(COleDateTime *current_time)

{

	BeginListAdjustment();

	COleDateTime	ct;

	if (!current_time)

		ct = COleDateTime::GetCurrentTime();

	else

		ct = *current_time;

	EventList_t	new_list;



	for (int x = 0; x < mEventList.size(); ++x)

		{

		COleDateTime n = mEventList[x].NextActivationTime();

		COleDateTimeSpan s = mEventList[x].Duration();

		if (n > ct)

			new_list.push_back(mEventList[x]);

		}

	mEventList = new_list;

	EndListAdjustment();

}





void		
CTimeShiftManager::EndListAdjustment(void)
{
	stack_MutexLock sml(&mMutex);
	assert(mAdjustingList);

	mAdjustingList = false;
	FigureNextEventOfInterest();
}

CTimeShiftManager::Action_t	

CTimeShiftManager::_Query(CTimeShiftEvent &event,COleDateTime *current_time,int warning_limit)

{

	Action_t	result = Nothing;



	stack_MutexLock sml(&mMutex);



	assert(!mAdjustingList);

	if (mNextEventOfInterest != -1)

		{

		event = mEventList[mNextEventOfInterest];

		COleDateTime ct;

		if (current_time)

			ct = *current_time;

		else

			ct = COleDateTime::GetCurrentTime();

		COleDateTimeSpan 	five_minutes(0,0,warning_limit,0);

		COleDateTimeSpan	one_minute(0,0,1,0);



		if (mRunning)

			{

			if (ct > (mStartTimeOfInterest + mDurationOfInterest))

				{

				result = StopCapture;	

				mRunning = false;

				mWarned = false;

				FigureNextEventOfInterest(current_time);

				}

			}

		else if (mWarned)

			{

			if (ct >= (mStartTimeOfInterest))

				{

				result = StartCapture;

				mRunning = true;

				}

			}

		else

			{

			if (ct >= (mStartTimeOfInterest))

				{

				result = StartCapture;

				mRunning = true;

				mWarned = true;

				}

			else if (ct >= (mStartTimeOfInterest - five_minutes))

				{

				result = Warning;

				mWarned = true;

				}

			}

		}

	return result;

}



CTimeShiftManager::Action_t	

CTimeShiftManager::Query(CTimeShiftEvent &event,COleDateTime current_time,int warning_limit)

{

	return _Query(event,&current_time,warning_limit);

}





CTimeShiftManager::Action_t	
CTimeShiftManager::Query(CTimeShiftEvent &event,int warning_limit)
{
	return _Query(event,0,warning_limit);

}

void		
CTimeShiftManager::AddEvent(const CTimeShiftEvent &event)
{
	stack_MutexLock sml(&mMutex);
	assert(mAdjustingList);

	mEventList.push_back(event);
}

void		
CTimeShiftManager::RemoveEvent(const CTimeShiftEvent &event)
{
	stack_MutexLock sml(&mMutex);
	assert(mAdjustingList);

	EventList_t::iterator i;
	for (i = mEventList.begin(); i != mEventList.end(); ++i)
		{
		if ((*i) == event)
			{
			mEventList.erase(i);
			break;
			}
		}
}


void

CTimeShiftManager::Clear(void)

{

	stack_MutexLock sml(&mMutex);

	assert(mAdjustingList);

	mEventList.erase(mEventList.begin(),mEventList.end());

}


void		
CTimeShiftManager::RemoveEvent(int index)
{
	stack_MutexLock sml(&mMutex);

	assert(mAdjustingList);

	EventList_t::iterator i;
	int x = 0;
	for (i = mEventList.begin(); i != mEventList.end(); ++i,++x)
		{
		if (x == index)
			{
			mEventList.erase(i);
			break;
			}
		}
}

CTimeShiftManager::EventList_t	
CTimeShiftManager::GetEventList(void)
{
	stack_MutexLock sml(&mMutex);

	return mEventList;
}


bool		

CTimeShiftManager::IsThereANextEvent(void)

{

	return (mNextEventOfInterest != -1);

}


