/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifdef	WIN32
#include "BuildClutFromPalette.h"
/*
BuildClutFromPalette

Builds the duck "double clut" from a palette. Uses an
oct-tree search.


*/

////////////////////////////////////////////

const int TREE_DEPTH = 5;	// bits per color in clut
const int MAX_COLOR = 32;	// 2^TREE_DEPTH
const int NUM_CHILDREN = 8;	// if it wasn't 8 it wouldn't be an
							// oct-tree, now would it?
const int COLOR_FIELDS = 3;	// r,g,b

// some types so we can loop
typedef unsigned char	RGBColor[COLOR_FIELDS];
typedef unsigned char	*RGBColorPtr;
typedef unsigned char	RGBRange[2];	// min/max
typedef RGBRange		RGBSpans[COLOR_FIELDS];

// oct-tree node
class ONode;
typedef ONode	*ONodePtr;

class ONode {
	ONodePtr	mChildren[NUM_CHILDREN];
	int			mIndex;
	int			mLoadData;
	int			mLeaf;
	RGBColor	mColors;
	RGBSpans	mSpans;

public:
	ONode(void)
		{
		mLoadData = 1;
		mLeaf = 0;
		for (int x = 0; x < NUM_CHILDREN; ++x)
			mChildren[x] = NULL;
		}
	~ONode(void)
		{
		for (int x = 0; x < NUM_CHILDREN; ++x)
			{
			if (mChildren[x]) delete mChildren[x];
			}
		}

private:

	////////////////////////////////////////////////
	// calc_dist
	//
	// calculate New york distance in three dimensions
	// from one color in the cube to another. You could
	// improve things by calculating real distance, but
	// the color space cube is so sparce it really doesn't
	// matter
	//
	// INPUT:
	//	c1 - source color
	//	c2 - destination color
	//
	// OUTPUT:
	//	New york distance
	///////////////////////////////////////////////////
	int	calc_dist(RGBColor c1,RGBColor c2)
		{
		return(abs(c1[0] - c2[0]) + abs(c1[1] - c2[1]) + abs(c1[2] - c2[2]));
		}

public:

	////////////////////////////////////////////////////
	// darkness
	//
	// takes an rgb value, encoded into a single int.
	// This routine has something to do with making sure
	// dark colors end up black. I'm not really sure why
	// this works, I just lifted it from the old Getpal
	// code. Seems to do the trick
	//
	// INPUT:
	//	rgb - rgb value encoded into a single int
	//
	// OUTPUT:
	//	some sort of darkness factor
	///////////////////////////////////////////////////////
	static int darkness(int rgb) 
		{
		// return max R, G, B
		int r,g,b;

		r=(rgb>>10)&0x1f;
		g=(rgb>>5)&0x1f;
		b=(rgb)&0x1f;

		if(g<r) g=r;
		if(g<b) g=b;
		return(g);
		}

	//////////////////////////////////////////////////
	// _FindClosestChild
	//
	// private method that recurses to find the leaf
	// closest to a given color
	//
	// INPUT:
	//	color	- color we are trying to match
	//	n		- pointer to found node
	//
	// OUTPUT:
	//	distance of color to node returned in parameter n
	//////////////////////////////////////////////////////
	int		_FindClosestChild(RGBColor color,ONode **n)
		{
			// am i a leaf?
			if (mLeaf)
				{
				// if I'm a leaf return my distance to the color
				*n = this;
				return calc_dist(color,mColors);
				}
			else
				{
				// I'm a node, return the distance to my closest
				// child
				int	mindist = 1000000;
				ONode *nn;
				for (int x = 0; x < NUM_CHILDREN; ++x)
					{
					if (mChildren[x])
						{
						int dist = mChildren[x]->_FindClosestChild(color,&nn);
						if (dist < mindist)
							{
							*n = nn;
							mindist = dist;
							}
						}
					}
				return mindist;
				}
		}

public:

	////////////////////////////////////////////////
	// FindClosestChild
	//
	// Starting with the current node, returns the
	// leaf closest to the color. It searchs the current
	// node and all it's children and their children
	// recursively
	//
	// INPUT:
	//	color	- color to locate
	//
	// OUTPUT:
	//	The closest node to the desired color
	////////////////////////////////////////////////////
	ONode*	FindClosestChild(RGBColor color)
		{
		ONode	*n;

			_FindClosestChild(color,&n);
			return n;
		}

	/////////////////////////////////////////////////////
	// AddNode
	//
	// Given a pointer to an rgb triple, and it's index in
	// the current palette, add it to the oct-tree
	//
	// INPUT:
	//	ppal	- pointer to rgb triple to add to tree
	//	index	- the palette index of the rgb color
	///////////////////////////////////////////////////////
	void AddNode(RGBColorPtr ppal,int index)
		{
		RGBSpans	spans = {{0,MAX_COLOR-1},{0,MAX_COLOR-1},{0,MAX_COLOR-1}};
		_AddNode(spans,ppal,index);
		}

private:

	//////////////////////////////////////////////////////////
	// _AddNode
	//
	// private method recurses down the tree to find a home
	// for the indicated color
	//
	// INPUT:
	//	spans	- an array of color spans that keeps track of
	//				where we are.
	//	ppal	- pointer to the rgb color to add
	//	index	- the color's index in the palette
	/////////////////////////////////////////////////////////////
	void _AddNode(RGBSpans &spans,unsigned char *ppal,int index)
		{
		int	min_slot = 0,max_slot = NUM_CHILDREN-1;
		
		// first time through, make note of my location in the tree
		if (mLoadData)
			{
			mLoadData = 0;
			for (int x = 0; x < COLOR_FIELDS; ++x)
				{
				mColors[x] = spans[x][0] + ((spans[x][1] - spans[x][0] + 1) / 2);
				mSpans[x][0] = spans[x][0];
				mSpans[x][1] = spans[x][1];
				}
			mIndex = index;
			}

		// See if this is the leaf, if so terminate recursion
		if ((spans[0][0] == spans[0][1]) && (spans[1][0] == spans[1][1]) && 
					(spans[2][0] == spans[2][1])) 
			{
			mLeaf = 1;
			return;
			}

		// calculate the next child to go to
		for (int x = 0;x < COLOR_FIELDS; ++x)
			{
			if (ppal[x] < mColors[x])
				{
				max_slot -= (max_slot - min_slot + 1) / 2;
				spans[x][1] -= (spans[x][1] - spans[x][0] + 1) / 2;
				}
			else
				{
				min_slot += (max_slot - min_slot + 1) / 2;
				spans[x][0] += (spans[x][1] - spans[x][0] + 1) / 2;
				}
			}

		// massive paranoid checks
		if (min_slot != max_slot) 
			exit(-1); // internal error
		if (min_slot < 0 || min_slot >= NUM_CHILDREN)
			exit(-1); // internal error

		// nobody here yet? create a new node
		if (!mChildren[min_slot]) 
			mChildren[min_slot] = new ONode;

		// keep goin'
		mChildren[min_slot]->_AddNode(spans,ppal,index);
		}

public:

	////////////////////////////////////////////////////////
	// MatchRGB
	//
	// returns the index of the best match for a given color
	//
	// INPUT:
	//	r,g,b	- the color of interest
	//	rigorous	- in the event of no match, how hard should
	//		we search. Use 1 for really hard, 0 for normal
	//
	// OUTPUT:
	//	The palette index of the best match to the color
	//////////////////////////////////////////////////////////
	int	MatchRGB(unsigned char r,unsigned char g,unsigned char b,int rigorous)
	{
	RGBColor	color = {r,g,b};
	RGBSpans	spans = {{0,MAX_COLOR - 1},{0,MAX_COLOR - 1},{0,MAX_COLOR-1}};

	return _MatchRGB(spans,color,rigorous);
	}

private:

	//////////////////////////////////////////////////
	// _MatchRGB
	//
	// The private recursive part of MatchRGB. Returns the
	// index in the palette of a given color
	//
	// INPUT:
	//	spans	- array of color spans that keeps track of
	//				where we are in the tree
	//	color	- The color we are searching for
	//	rigorous- how hard should we search when a match
	//				is not found. 0 is normal, 1 is hard
	////////////////////////////////////////////////////////////
	int _MatchRGB(RGBSpans &spans,RGBColor color,int rigorous)
		{
		int	min_slot = 0,max_slot = NUM_CHILDREN-1;
		int cut;

		// If this is a leaf then terminate recursion
		if ((spans[0][0] == spans[0][1]) && (spans[1][0] == spans[1][1]) && 
			(spans[2][0] == spans[2][1])) 
			{
			return mIndex;
			}

		// figure out which child to go to
		for (int x = 0; x < COLOR_FIELDS; ++x)
			{
			cut = spans[x][0] + ((spans[x][1] - spans[x][0] + 1) / 2);
			if (color[x] < cut)
				{
				max_slot -= (max_slot - min_slot + 1) / 2;
				spans[x][1] -= (spans[x][1] - spans[x][0] + 1) / 2;
				}
			else
				{
				min_slot += (max_slot - min_slot + 1) / 2;
				spans[x][0] += (spans[x][1] - spans[x][0] + 1) / 2;
				}
			}

		// massive paranoia checks
		if (min_slot != max_slot) 
			exit(-1); // internal error
		if (min_slot < 0 || min_slot >= NUM_CHILDREN)
			exit(-1);

		// If there is no child, then we must approximate
		// the color
		if (!mChildren[min_slot]) 
			{
			if (rigorous)	// rigorous approximation - could be slow
				return FindClosestChild(color)->mIndex;

			// color not available, take a guess. By finding
			// the node with the closest centroid to our color
			// and continue down that path
			int	mindist = 1000000;
			int minindex;

			for (int y = 0; y < NUM_CHILDREN; ++y)
				{
				if (mChildren[y])
					{
					int dist = calc_dist(color,mChildren[y]->mColors);
					if (dist < mindist)
						{
						mindist = dist;
						minindex = y;
						}
					}
				}
			min_slot = minindex;
			for (x = 0; x < COLOR_FIELDS; ++x)
				{
				spans[x][0] = mChildren[minindex]->mSpans[x][0];
				spans[x][1] = mChildren[minindex]->mSpans[x][1];
				}

			}
		
		// Psychotically paranoid checks
		if (min_slot < 0 || min_slot >= NUM_CHILDREN)
			exit(-1); // internal error
		if (!mChildren[min_slot]) 
			exit(-1); // internal error

		// keep going
		return mChildren[min_slot]->_MatchRGB(spans,color,rigorous);
		}				
};


/////////////////////////////////////////////////////////////
// BuildClutFromPalette
//
// Given a palette, builds a color lookup table.
//
// INPUT:
//	clut - the resulting color lookup table. This is a table
//		of two byte entries that correspond to the clut
//		that is parsed by the duck player.
//	clutsize - size of the table (in entries not bytes). 236 is typical
//	pal	- the input palette. This is an array of rgb triples
//	palsize - size of the palette in entries (not bytes)
//	rigorous - if 1 then a slightly more intense search is
//		done on color match failure. Frankly, it doesn't
//		seem to make any difference, so I always use 0
//
// WARNING !!!! - BOTH the clut and pal data will be modified
// by this procedure
//
/////////////////////////////////////////////////////////////////
void BuildClutFromPalette(WORD *clut,int clutsize,unsigned char *pal,
						  int palsize,int rigorous)
{
unsigned char *pal2;
ONode	*tree_root;


	// build octree
	pal2 = pal;
	tree_root = new ONode;

	// convert palette to 5 bit
	for (int x = 0; x < palsize; ++x,pal2+=3)
		{
		pal2[0] = pal2[0] >> 3;
		pal2[1] = pal2[1] >> 3;
		pal2[2] = pal2[2] >> 3;
		tree_root->AddNode(pal2,x);
		}

	// build clut
	for (x = 0; x < clutsize; ++x)
		{
		unsigned char r,g,b;

		r = (x >> 10) & 0x1f;
		g = (x >> 5) & 0x1f;
		b = x & 0x1f;
		int	index = tree_root->MatchRGB(r,g,b,rigorous);
		clut[x] = index+1;
		if (ONode::darkness(x) >= 6)
			{
			r = ((3*r)-(pal[(index*3)+0]))/2;
			g = ((3*g)-(pal[(index*3)+1]))/2;
			b = ((3*b)-(pal[(index*3)+2]))/2;
			index = tree_root->MatchRGB(r,g,b,rigorous);
			clut[x] += (index+1) << 8;
			}
		else
			{
			clut[x] += (clut[x] << 8);
			}
		}

	delete tree_root;
}
#endif

