/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CCircularBuffer.h>
#include	<CLogger.h>
#include	<stack_aids.h>

/******************************************************************************
CCircularBuffer

	Implements a circular buffer with wraparound data and index space. 
	It is used
	to mimick the HFB indexed buffering scheme on a stream by stream basis. We
	create a finite set of indexes that wraps around in order to emulate the
	HFB index scheme in a situation where there are no indexes

///////// WARNING... THIS BUFFER ONLY PROPERLY SUPPORTS ONE READER AND
///////// ONE WRITER

******************************************************************************/

/////////////////////////////////////////////////////////////////////////
// CCircularBuffer
//
// Constructor
//
// INPUT:
//	pool_size	- size of data space for buffer
//	num_indicies- number of indicies to use
///////////////////////////////////////////////////////////////////////////
CCircularBuffer::CCircularBuffer(size_t pool_size,int num_indicies)
{
	mMemoryPool = NULL;
	mIndicies = NULL;

	mMemoryPool	= new char[pool_size];
	mEndOfPool 	= mMemoryPool + pool_size;
	mWritePtr 	= mMemoryPool;
	mBlockPtr 	= mMemoryPool;
	mDynamicEndOfPool = mMemoryPool;

	mIndicies = new CB_Index[num_indicies];
	for (int x = 0; x < num_indicies; ++x)
		{
		mIndicies[x].mBusy = 0;
		mIndicies[x].mPtr = NULL;
		}
	mMinIndex 			= 0;
	mCurrentReadIndex 	= -1;
	mCurrentWriteIndex	= 0;
	mMaxIndex 			= num_indicies - 1;
	mIgnoreOverflow		= 0;
	mChunksInBuffer = 0;
}

////////////////////////////////////////////////////////////////
// ~CCircularBuffer
//
// Destructor
/////////////////////////////////////////////////////////////////
CCircularBuffer::~CCircularBuffer(void)
{
	delete [] mMemoryPool;
	mMemoryPool = NULL;

	delete [] mIndicies;
	mIndicies = NULL;
}

////////////////////////////////////////////////////////////
// Clear
//
// reset the circular buffer to it's initial empty state
////////////////////////////////////////////////////////////
void
CCircularBuffer::Clear(void)
{
	stack_MutexLock sml(&mMutex);

	mWritePtr 	= mMemoryPool;
	mBlockPtr 	= mMemoryPool;
	mDynamicEndOfPool = mMemoryPool;

	int num_indicies = mMaxIndex + 1;

	for (int x = 0; x < num_indicies; ++x)
		{
		mIndicies[x].mBusy = 0;
		mIndicies[x].mPtr = NULL;
		}
	mCurrentReadIndex 	= -1;
	mCurrentWriteIndex	= 0;
	mChunksInBuffer = 0;

	mWaitingToWrite.Signal();
}

/////////////////////////////////////////////////////////////////////
// WaitForWriteSpace
//
// This is called by a thread that wants to add new data to the buffer.
// You call this with the amount of space you need. 
//
// INPUT:
//	siz	- amount of data space requested
//
// OUTPUT:
//	Returns a pointer to the requested space or NULL and blocks depending
//	on the state of the IgnoreOverflow flag. If IgnoreOverflow is true
//	then a NULL pointer is returned if the buffer is full. If IgnoreOverflow
//	is false then the method blocks until a reader thread frees up enough
//	space. If IgnoreOverflow is false the method will never return NULL.
//////////////////////////////////////////////////////////////////////////// 
char *
CCircularBuffer::WaitForWriteSpace(size_t siz)
{
	stack_MutexLock sml(&mMutex);

	int		blocked;

#ifdef	ALIGN_FACTOR
	siz = (siz + ALIGN_FACTOR) & ~((size_t)(ALIGN_FACTOR-1));
#endif

	// wait for our space
	do 
		{
		blocked = 0;

		if (mIndicies[mCurrentWriteIndex].mBusy)
			{
			blocked = 1;
			}
		else if (mBlockPtr > mWritePtr)
			{
			if ((mBlockPtr - mWritePtr - 1) < ((long)siz))
				blocked = 1;
			}
		else
			{
			if (mEndOfPool - mWritePtr < ((long)siz))
				{
				if (mBlockPtr == mMemoryPool)
					{
					blocked = 1;
					}
				else
					{
					mDynamicEndOfPool = mWritePtr;
					mWritePtr = mMemoryPool;
					if (mBlockPtr - mWritePtr - 1 < ((long)siz))
						blocked = 1;
					}
				}
			}

		if (blocked)
			{
			if (!mIgnoreOverflow)
				mWaitingToWrite.Wait(mMutex);
			gTraceLog->SendMessage("Circle block");
			}
		}
	while((!mIgnoreOverflow) && blocked);
	
	return (blocked ? ((char *)NULL) : mWritePtr);
}

///////////////////////////////////////////////////////////////////////////
// AdvanceWriteMark
//
// This is called by a writer thread after it is done with the pointer
// returned by WaitForWriteSpace(). You MUST pass it the same size
// as in WaitForWriteSpace(). This method makes the data available to reader
// threads.
//
// INPUT:
//	siz	- size of data to make available to reader. MUST be the same value
//			passed to WaitForWriteSpace()
//////////////////////////////////////////////////////////////////////////////
void
CCircularBuffer::AdvanceWriteMark(size_t siz)
{
	stack_MutexLock sml(&mMutex);

#ifdef	ALIGN_FACTOR
	mIndicies[mCurrentWriteIndex].mReportedLength = siz;
	siz = (siz + ALIGN_FACTOR) & ~((size_t)(ALIGN_FACTOR-1));
#endif
	mIndicies[mCurrentWriteIndex].mBusy = 1;
	mIndicies[mCurrentWriteIndex].mPtr = mWritePtr;
	mIndicies[mCurrentWriteIndex].mLen = siz;

	mCurrentWriteIndex++;
	if (mCurrentWriteIndex > mMaxIndex)
		mCurrentWriteIndex = mMinIndex;

	mWritePtr += siz;

	if (mWritePtr > mDynamicEndOfPool)
		mDynamicEndOfPool = mWritePtr;

	mChunksInBuffer++;

	mWaitingToRead.Signal();
}

////////////////////////////////////////////////////////////////////
// FetchIndex
//
// Does the first part of the HFB_GetStreamingData emulation. A reader
// thread calls this to get an index value. This index can then be
// used to extract the data with GetDataFromIndex(). when you are
// done with the index you call FreeIndex() which makes the data space
// available to the writer thread
//
// INPUT:
//	direction	- Duck direction as in the call to HFB_GetStreamingData
//	val			- Duck direction count as in the call to HFB_GetStreamingData
//
// OUTPUT:
//	index value or -1 if error or -2 if no data available (just like
//		HFB_GetStreamingData.
//
///////////////////////////////////////////////////////////////////////////
int
CCircularBuffer::FetchIndex(int direction,int val)
{
	stack_MutexLock sml(&mMutex);

	int	result;
	int	tmp,loop;

	tmp = mCurrentReadIndex;

	loop = (val || (!direction));
	while (val || loop)
		{
		loop = 0;
		switch(direction)
			{
			case -1:	// backwards
				tmp--;
				if (tmp < mMinIndex)
					tmp = mMaxIndex;
				break;
			case 0:		// absolute
				tmp = val;
				val = 1; // force exit
				break;
			case 1:		// forward
				tmp++;
				if (tmp > mMaxIndex)
					tmp = mMinIndex;				
				break;
			}


		if (!mIndicies[tmp].mBusy)
			{
			result = -2;
			val = 0;
			}
		else
			{
			result = tmp;
			val--;
			}
		}

	if (result >= 0)
		mCurrentReadIndex = tmp;

	return result;
}

/////////////////////////////////////////////////////////////////////////
// GetDataFromIndex
//
// Assuming you already have a valid index from FetchIndex() you can
// use this routine to get at the data
//
// INPUT:
//	index	- the index whose data you want
//
// OUTPUT:
//	data	- the pointer to the data
//	len		- the size of the data
//////////////////////////////////////////////////////////////////////////
void
CCircularBuffer::GetDataFromIndex(int index,char **data,long *len)
{
	stack_MutexLock sml(&mMutex);

	if (index == -1) // for HFB_getdataPos emulation in CNetworkContext
		index = mCurrentReadIndex;
	
	if (index == -1)
		index = 0; // for HFB_getdataPos emulation. The buffer might have
					// just been cleared and mCurrentReadIndex is -1
#ifdef	ALIGN_FACTOR
	*len = mIndicies[index].mReportedLength;
#else
	*len = mIndicies[index].mLen;
#endif
	*data = mIndicies[index].mPtr;
}

//////////////////////////////////////////////////////////////////////
// FreeIndex
//
// Call this when you are completely done with an index. This returns
// the data space to the pool and makes it available to the writing
// thread
//
// INPUT:
//	index	- index to free
///////////////////////////////////////////////////////////////////////
void
CCircularBuffer::FreeIndex(int index)
{
	stack_MutexLock sml(&mMutex);

	if (mBlockPtr >= mDynamicEndOfPool && mBlockPtr != mWritePtr)
		mBlockPtr = mMemoryPool;

	if (mIndicies[index].mBusy)
		{
		mChunksInBuffer--;
		mIndicies[index].mBusy = 0;
		if (mIndicies[index].mPtr == mBlockPtr)
			{
			while (mIndicies[index].mBusy == 0 && 
												mIndicies[index].mPtr == mBlockPtr)
				{
				mBlockPtr += mIndicies[index].mLen;
				mIndicies[index].mPtr = NULL;
				index++;
				if (index > mMaxIndex)
					index = mMinIndex;
				if (mBlockPtr >= mDynamicEndOfPool && mBlockPtr != mWritePtr)
					mBlockPtr = mMemoryPool;
				}

			mWaitingToWrite.Signal();
			}
		}
}

//////////////////////////////////////////////////////////////////////////////
// ReadyToRoll
//
// We need to guess when this buffer is ready. We pass in a percent threshold
//
// INPUT:
//	percent	- percentage beyond which we are ready
//
// OUTPUT:
//	1 if ready else 0
/////////////////////////////////////////////////////////////////////////////
int
CCircularBuffer::ReadyToRoll(long percent)
{
	if ((mCurrentWriteIndex - mMinIndex) >= (((mMaxIndex - mMinIndex) * percent) / 100l))
		return 1;

	if ((mWritePtr - mMemoryPool) >= (((mEndOfPool - mMemoryPool) * percent) / 100l))
		return 1;

	return 0;
}

////////////////////////////////////////////////////////////////////////////
// GetPercentFull
//
// Return buffer percent full
//
// OUTPUT:
//	percent full
////////////////////////////////////////////////////////////////////////////
int
CCircularBuffer::GetPercentFull(void)
{
	long	amount_remaining;


	if (mBlockPtr > mWritePtr)
		{
		amount_remaining = mBlockPtr - mWritePtr;
		}
	else
		{
		amount_remaining = mEndOfPool - mWritePtr;
		amount_remaining += mBlockPtr - mMemoryPool;
		}

	return (100 - ((amount_remaining * 100) / (mEndOfPool - mMemoryPool)));
}

int
CCircularBuffer::GetChunksInBuffer(void)
{
	return mChunksInBuffer;
}

long
CCircularBuffer::GetBytesInBuffer(void)
{
	long	amount_remaining;


	if (mBlockPtr > mWritePtr)
		{
		amount_remaining = mBlockPtr - mWritePtr;
		}
	else
		{
		amount_remaining = mEndOfPool - mWritePtr;
		amount_remaining += mBlockPtr - mMemoryPool;
		}

	return mEndOfPool - mMemoryPool - amount_remaining;
}


////////////////////////////////////////////////////////////////////
// SetOverflowIgnore
//
// set the value of the ignore overflow flag for writer thread
//
// INPUT:
//	val	 - new value for ignoreoverflow flag (1 or 0)
//////////////////////////////////////////////////////////////////////
void	
CCircularBuffer::SetOverflowIgnore(int val)
{
	stack_MutexLock sml(&mMutex);

	mIgnoreOverflow = val;
	mWaitingToWrite.Signal();
}










