/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CCThread.h>

/*******************************************
CCThread

Class represents a thread of execution
********************************************/

//////////////////////////////////////////////
// CCThread
//
// Constructor. Does not activate thread
//////////////////////////////////////////////
CCThread::CCThread(void)
{
#ifdef	WIN32
	mThreadID = 0;
#elif __POWERPC__
	mThreadID = kNoThreadID;
#endif
	mRunning = 0;
}

//////////////////////////////////////////////
// ~CCThread
//
// Destructor. Kills thread if still running
///////////////////////////////////////////////
CCThread::~CCThread(void)
{
	assert(!mRunning); // I'm not sure if we should allow this yet
	if (mRunning)
		{
		Cancel();
		throw DtorException("CCThread::!CCThread destroyed while running",0);
		}
#ifdef	WIN32
	if (mThreadID != NULL && ((long)mThreadID) != -1)
		CloseHandle((void *)mThreadID);
#elif __POWERPC_
	long	result;
	if (mThreadID != kNoThreadID)
		::DisposeThread(mThreadID,&result,FALSE);
	mThreadID = kNoThreadID;
#endif
}

#ifdef	WIN32
unsigned int __stdcall thread_start(void *arg)
{
	((CCThread *)(arg))->_Go();
	return 0;
}
#elif __POWERPC__
pascal void *thread_start(void *arg)
{
	CCThread* volatile thread = reinterpret_cast<CCThread *>(arg);
	thread->_Go();
	return NULL;
}

#else
void *thread_start(void *arg)
{
	((CCThread *)(arg))->_Go();
	return NULL;
}
#endif

//////////////////////////////////////////
// Start
//
// Call to begin threads execution.
/////////////////////////////////////////
void
CCThread::Start(void)
{
#ifdef	WIN32
	unsigned	int	id;

	if (mThreadID != NULL && ((long)mThreadID) != -1)
		CloseHandle((void *)mThreadID);
	mThreadID = _beginthreadex(NULL,0,thread_start,this,0,&id);
	if (mThreadID == NULL || ((long)mThreadID) == -1)
		throw StartException("CCThread::Start",(int)GetLastError());
#elif __POWERPC__
	long result;
	if (mThreadID != kNoThreadID)
		::DisposeThread(mThreadID,&result,FALSE);
	::NewThread(kCooperativeThread,thread_start,(void *)this,0,kNewSuspend,nil,
				&mThreadID);
	if (mThreadID == kNoThreadID)
		throw StartException("CCThread::Start error",mThreadID);
	OSErr err = ::SetThreadState(mThreadID,kReadyThreadState,kNoThreadID);
	if (err != noErr)
		throw StartException("CCThread::Start",err);
#else
	int 	err;

	if (err=pthread_create(&mThreadID,NULL,thread_start,this))
		throw StartException("CCThread::Start",err);
#endif

	mRunning = 1;
}

///////////////////////////////////////////
// _Go
//
// Private method that invokes run. Used
// so running flag is automatically managed
/////////////////////////////////////////////
void
CCThread::_Go(void)
{
	Run();
	mRunning = 0;
}


//////////////////////////////////////////////////////
// Cancel
//
// Cancel this thread
//////////////////////////////////////////////////////
void
CCThread::Cancel(void)
{
#ifdef	WIN32
	if (!TerminateThread((void *)mThreadID,(DWORD)-1))
		throw CancelException("CCThread::Cancel",(int)GetLastError());

#elif __POWERPC__
	assert(mThreadID);
	OSErr err = ::SetThreadState(mThreadID,kStoppedThreadState,kNoThreadID);
	if (err != noErr)
		throw CancelException("CCThread::Cancel",err);
#else
	int err;

	if (err = pthread_cancel(mThreadID))
		throw CancelException("CCThread Cancel",err);

#endif
	mRunning = 0;
}

//////////////////////////////////////////////////
// ForceExit
//
// Force the thread of execution to exit
///////////////////////////////////////////////////
void
CCThread::ForceExit(void)
{
#ifdef	WIN32
	_endthreadex(0);
#elif __POWERPC__
	ThreadID id;
	::GetCurrentThread(&id);
	OSErr err = ::SetThreadState(id,kStoppedThreadState,kNoThreadID);
#else
	pthread_exit(NULL);
#endif
}

///////////////////////////////////////////////////////////////////////////////
// Wait for thread to exit
//
///////////////////////////////////////////////////////////////////////////////
void
CCThread::Join(void)
{

#ifdef	WIN32
	WaitForSingleObject((void *)mThreadID,INFINITE);
#elif __POWERPC__
	while (mRunning)
		{
		OSErr err = ::YieldToAnyThread();
		if (err != noErr)
			throw JoinException("CCThread::Join",err);
		}
#else
	int	err;
	void *ptr;

	err = pthread_join(mThreadID,&ptr);
	// join can return errors if thread has already exited. So
	// fuck it, and ignore the errors
	//if (err && err != ESRCH)
	//	throw JoinException("CCThread::Join",err);
#endif

}

//////////////////////////////////////////////////////////////////////////////
// Yield
//
// yields the CALLING thread. Because of this the method is static as a 
// convention
//////////////////////////////////////////////////////////////////////////////
void
CCThread::Yield(void)
{
#ifdef	WIN32
	Sleep(0);
#elif __POWERPC__
	OSErr err = ::YieldToAnyThread();
	if (err != noErr)
		throw YieldException("CCThread::Yield",err);
#else
	sched_yield();
#endif
}

///////////////////////////////////////////////////////////////////////////////
// Kill
//
// sends a signal to this thread
//
// INPUT:
//	signal	- the unix style signal to send
//
///////////////////////////////////////////////////////////////////////////////
void
CCThread::Kill(int signal)
{
#ifdef	WIN32
	Cancel();
#elif __POWERPC__
	Cancel();
#else
	int err;

	err = pthread_kill(mThreadID,signal);
	// kill can return errors if a thread has already exited
	// it causes us great problems, so we just ignore it
	//	if (err && err != ESVAL)
	//		throw KillException("CCThread::Kill",err);
#endif
}

///////////////////////////////////////////////////////////////////////////////
// IsRunning
//
// Is the thread running
//
// OUTPUT:
//	true if thread is still running
///////////////////////////////////////////////////////////////////////////////
int
CCThread::IsRunning(void) const
{
	return mRunning;
}


CCThread_ID_Type
CCThread::ReturnNativePtr(void) const
{
	return mThreadID;
}

CCThread_ID_Type
CCThread::Self(void)
{
#ifdef	WIN32
	return (unsigned long)GetCurrentThreadId();
#elif __POWERPC__
	ThreadID	id;
	::GetCurrentThread(&id);
	return id;
#else
	return pthread_self();
#endif
}


void
CCThread::Dump(ostream &o,const char *indent)
{
	o << indent << "CCThread: " << endl
		<< indent << "   mRunning:" << mRunning << endl
		<< indent << "   mThreadID:" << mThreadID << endl;
}

ostream& operator<<(ostream &o,CCThread &t)
{
	t.Dump(o,"");
	return o;
}

ostream& operator<<(ostream &o,CCThread *t)
{
	if (t)
		t->Dump(o,"");
	else
		o << " NULL " << endl;
	return o;
}

const char *
CCThread::Name(void)
{
	return "CCThread";
}

#ifdef	IRIX
#include	<exception.h>
// Irix multithread exception kludge
long tif(void)
{
	return CCThread::Self();
}

class IrixExceptionHack
{
public: 
	IrixExceptionHack(void)	
		{
		set_thread_id_function(tif);
		}
};

IrixExceptionHack ieh;

#endif





