/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CConditionVariable.h>
#include	<nmr_string.h>
#ifdef	__POWERPC__
#include	<system_ticks.h>
#include	<CCThread.h>
#endif

/******************************************************
CConditionVariable

A trivial encapsulation of pthread condition variables.
Exceptions are thrown instead of returning error codes
*********************************************************/

CConditionVariable::CConditionVariable(void)
{
#ifdef	WIN32
	mCondition = CreateEvent(0,TRUE,FALSE,NULL);
	if (mCondition == NULL)
		throw CtorException("CConditionVariable::CConditionVariable",
														(int)GetLastError());
#elif __POWERPC__
	mCondition = 0;
#else
	int	err;

	if (err=pthread_cond_init(&mCondition,NULL))
		throw CtorException("CConditionVariable::CConditionVariable",err);
#endif
}

CConditionVariable::~CConditionVariable(void)
{
#ifdef	WIN32
	if (mCondition != NULL)
		CloseHandle(mCondition);
#elif __POWERPC__
#else
	int	err;

	if (err=pthread_cond_destroy(&mCondition))
		throw DtorException("CConditionVariable::~CConditionVariable",err);
#endif
}

void
CConditionVariable::Wait(CCMutex &mutex)
{
#ifdef	WIN32
	DWORD	err;
	
	mutex.Unlock();
	err = WaitForSingleObject(mCondition,INFINITE);
	if (err == WAIT_FAILED)
		throw WaitException("CConditionVariable::Wait WAIT_FAILED",err);
	mutex.Lock();
	ResetEvent(mCondition);

#elif __POWERPC__
	mutex.Unlock();
	while (!mCondition)
		CCThread::Yield();
	mCondition = 0;
	mutex.Lock();
#else
	int	err;

	if (err=pthread_cond_wait(&mCondition,&(mutex.mMutex)))
		throw WaitException("CConditionVariable::Wait",err);
#endif
}

void	
CConditionVariable::WaitInterval(CCMutex &mutex,int sec)
{
/* 
   note: timeout is not really an error. It can be totally normal
   behavior. Therefore, we don't want to throw a data message
   with the exception because copying the data message will slow us
   down.
*/

#ifdef	WIN32
	DWORD	err;
	
	mutex.Unlock();
	err = WaitForSingleObject(mCondition,sec*1000);
	if (err == WAIT_TIMEOUT)
		{
		mutex.Lock();
		throw TimeoutException();
		}
	if (err == WAIT_FAILED)
		throw WaitException("CConditionVariable::WaitInterval failed",err);

	mutex.Lock();
	ResetEvent(mCondition);

#elif __POWERPC__
	U32	start_time = timeGetTime();
	mutex.Unlock();
	while (!mCondition)
		{
		U32 elapsed = timeGetTime() - start_time;
		if (elapsed > (sec * 1000))
			{
			mutex.Lock();
			throw TimeoutException();
			}
		CCThread::Yield();
		}
	mCondition = 0;
	mutex.Lock();

#else
	int	err;

	struct timespec to;

	to.tv_sec = time(NULL) + sec;
	to.tv_nsec = 0;
	if (err=pthread_cond_timedwait(&mCondition,&(mutex.mMutex),&to))
		{
		if (err == ETIME)
			throw TimeoutException();
		throw WaitException("CConditionVariable::WaitInterval",err);
		}
#endif
}


void
CConditionVariable::Signal(void)
{
#ifdef	WIN32
	if (!SetEvent(mCondition))
		throw SignalException("CConditionVariable::Signal",
														(int)GetLastError());

#elif __POWERPC__
	mCondition = 1;
	CCThread::Yield();
#else
	int	err;

	if (err=pthread_cond_signal(&mCondition))
		throw SignalException("CConditionVariable::Signal",err);
#endif
}

void
CConditionVariable::Broadcast(void)
{
#ifdef	WIN32
	if (!PulseEvent(mCondition))
		throw BroadcastException("CConditionVariable::Broadcast",
														(int)GetLastError());
#elif __POWERPC__
	mCondition = 1;
	CCThread::Yield();
#else
	int	err;

	if (err=pthread_cond_broadcast(&mCondition))
		throw BroadcastException("CConditionVariable::Broadcast",err);
#endif
}

void
CConditionVariable::Dump(ostream &o,const char *indent) const
{
	o << indent << "CConditionVariable" << endl
		<< indent << "   " << (&mCondition) << endl;
}

ostream& operator<<(ostream &o,CConditionVariable &c)
{
	c.Dump(o,"");
	return o;
}

ostream& operator<<(ostream &o,CConditionVariable *c)
{
	if (c)
		c->Dump(o,"");
	else
		o << " NULL " << endl;
	return o;
}

