/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CFsm.h>
#include	<CLogger.h>
#include	<safe_ostrstream.h>

/*
	Finite State Machine
*/

///////////////////////////////////////////////////////////////////////////////
// CFsm
//
// Constructor
//
// INPUT:
//	max_queue_size	- maximum size of message queue
///////////////////////////////////////////////////////////////////////////////
CFsm::CFsm(int max_queue_size)
	: CMos(max_queue_size)
{
	mStateTable = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// ~CFsm
//
// Destructor
///////////////////////////////////////////////////////////////////////////////
CFsm::~CFsm(void)
{
	ClearStateTable();
}

///////////////////////////////////////////////////////////////////////////////
// ClearStateTable
//
// Clears the state table
///////////////////////////////////////////////////////////////////////////////
void
CFsm::ClearStateTable(void)
{
	if (mStateTable != NULL)
		{
		for (int x = 0; x <= mMaxState; ++x)
			{
			delete [] mStateTable[x];
			}
		delete [] mStateTable;
		}
}

///////////////////////////////////////////////////////////////////////////////
// AttachStateTable
//
// Attach a state table to the fsm. The table is copied into the fsm.
//
// INPUT:
//	table		- state table
//	num_entries	- number of entries in state able
///////////////////////////////////////////////////////////////////////////////
void
CFsm::AttachStateTable(state_table_entry_t *table,int num_entries)
{
	int	x;

	ClearStateTable();
	mMaxEvent = mMaxState = 0;

	// find the maximum state and event
	for (x = 0; x < num_entries; ++x)
		{
		if (table[x].mCurrentState > mMaxState)
			mMaxState = table[x].mCurrentState;
		if (table[x].mNextState > mMaxState)
			mMaxState = table[x].mNextState;
		if (table[x].mTransition > mMaxEvent)
			mMaxEvent = table[x].mTransition;
		}


	// construct table and initialize to null callback and
	// no state change
	mStateTable = new state_stub_ptr_t[mMaxState+1];
	for (x = 0; x <= mMaxState; ++x)
		{
		mStateTable[x] = new state_stub_t[mMaxEvent+1];
		for (int y = 0; y <= mMaxEvent; ++y)
			{
			(mStateTable[x])[y].mNextState = x;
			(mStateTable[x])[y].mAction = &CFsm::Null_Callback; 
			}
		}

	for (x = 0; x < num_entries; ++x)
		{
		state_stub_t	*state;

		state = mStateTable[table[x].mCurrentState];
		state[table[x].mTransition].mNextState	= table[x].mNextState;
		state[table[x].mTransition].mAction 	= table[x].mAction;
		}
	mCurrentState = 0;
}

///////////////////////////////////////////////////////////////////////////////
// MessageHandler
//
// Dispatches events send to the fsm through the state table. You generally
// shouldn't override this, since the state table should drive the entire
// fsm. The appropriate callback for the current state and the event
// code is envoked, then the state is advanced. EventOutOfBounds() is called
// if the event is beyond the scope of the state table. If it is in bounds
// but not set then Null_Callback() is called
//
// INPUT:
//	mess	- the message to handle
///////////////////////////////////////////////////////////////////////////////
void	
CFsm::MessageHandler(CMosMessage *mess)
	{
	fsm_callback_t	callback;
	state_stub_t	*stub;
	int				event;

		stub = (mStateTable[mCurrentState]);

		event = mess->GetCode();
		if (event  > mMaxEvent || event < 0)
			{
			EventOutOfBounds(mess);
			}
		else 
			{
			callback = stub[mess->GetCode()].mAction;
			(this->*callback)(mess);
			mCurrentState = stub[mess->GetCode()].mNextState;
			}
	}

///////////////////////////////////////////////////////////////////////////////
// EventOutOfBounds
//
// Called if an event is beyond the scope of the state table. override this
// if you want to handle this type of condition.
//
// INPUT:
//	mess	- the message that caused the error.
///////////////////////////////////////////////////////////////////////////////
void
CFsm::EventOutOfBounds(CMosMessage *mess)
{
}

///////////////////////////////////////////////////////////////////////////////
// Null_Callback
//
// Called if no method has been installed in the state table for the
// event. Override if you want to handle that situation.
//
// INPUT:
//	message - the message that did not have a handler
///////////////////////////////////////////////////////////////////////////////
void		
CFsm::Null_Callback(CMosMessage *message)
{
}

const char *
CFsm::Name(void)
{
	return "CFsm";
}

void
CFsm::Dump(ostream &o,const char *indent)
{
	o << indent << "CFsm:" << endl
		<< indent << "   mCurrentState: " << mCurrentState << endl
		<< indent << "   mMaxEvent: " << mMaxEvent << endl
		<< indent << "   mMaxState: " << mMaxState << endl;
	nmr_string	s(indent);
	s += "   ";
	CMos::Dump(o,s);
}

///////////////////////////////////////////////////////////////////////////////
// TraceCallback
//
// Send the name of the callback function and object to the trace log
//
// INPUT:
//	name - the name of the function
///////////////////////////////////////////////////////////////////////////////
void
CFsm::TraceCallback(const char *name)
{
	safe_ostrstream	o;

	o << Name() << "::" << name;
	gTraceLog->SendMessage(o.str());
}

