/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CMos.h>
#include	<CError.h>
#include	<CMosMessage.h>
#include	<CMosMessageType.h>
#include	<CLogger.h>
#include	<CFifo.h>
#include	<signal.h>
#include	<safe_ostrstream.h>
#include	<stack_aids.h>
#include	<sduplicate.h>
#include	<map>
using namespace std;

/******************************************************************************
CMos
	message oriented server. A thread that accepts and processes messsages
	of the type CMosMessage
******************************************************************************/

#ifdef	NEEDS_STL_ALLOCATOR
	typedef map<CMos*,CCThread_ID_Type,less<CMos*>,
								allocator<CCThread_ID_Type>  >	Mos_Table_t;
#else
	typedef map<CMos*,CCThread_ID_Type,less<CMos*> >	Mos_Table_t;
#endif

// these two cant be class variables because WIN32 gets very confused
// in CVidcapCore because of the stl stuff.
static Mos_Table_t	sMosTable;
static CCMutex		sMosTableMutex;

int					CMos::sSigKill = 0;
int					CMos::sSignal = 0;

//////////////////////////////////////////////////////////////////////////
// CMos
//
// Constructor. Inits message queue
//
//	INPUT:
//		maxMessages	- size of message queue
///////////////////////////////////////////////////////////////////////////
CMos::CMos(int maxMessages)
	: mMessageQueue(maxMessages)
{
	mTerminate = 0;
	mNoMoreMessages = 0;
}

/////////////////////////////////////////////////////////////////////////
// ~CMos
//
// Destructor
//////////////////////////////////////////////////////////////////////
CMos::~CMos(void)
{
	mNoMoreMutex.Lock();
	mNoMoreMessages = 1;
	mNoMoreMutex.Unlock();

	RemoveMosFromTable();	// just in case

	assert(!InMosList());
	
	int nofail;
	do
		{
		CMosMessage *mess = mMessageQueue.GetCanFail(nofail);
		if (nofail)
			delete mess;	
		}
	while (nofail);
}

///////////////////////////////////////////////////////////////////////////////
// AddMosToTable
//
// Add a mos to the table of running moses. This table is used to
// dispatch signals as messages
//
//////////////////////////////////////////////////////////////////////////////
void
CMos::AddMosToTable(void)
{
	stack_MutexLock sml(&sMosTableMutex);

	sMosTable[this] = mThreadID;
}

int
CMos::InMosList(void)
{
	stack_MutexLock sml(&sMosTableMutex);

	return (sMosTable.find(this) != sMosTable.end());
}

//////////////////////////////////////////////////////////////////////////////
// RemoveMosFromTable
//
// Remove a mos from the running table
//////////////////////////////////////////////////////////////////////////////
void
CMos::RemoveMosFromTable(void)
{
	stack_MutexLock sml(&sMosTableMutex);

	sMosTable.erase(this);
}
	
void	
CMos::Start(void)
{
	SendMessage(new CMosMessage(CMOS_SYS_INIT_OK));
	CCThread::Start();
}


///////////////////////////////////////////////////////////////////////////
// Run
//
// Override of CCThread's virtual method.
// Waits on the message queue and invokes MessageHandler() when a message is 
// available. Errors are handled by sending a message of type 
// CMosMessage::CMOS_SYS_INTERNAL_ERROR to one's self
////////////////////////////////////////////////////////////////////////////
void
CMos::Run(void)
{

	AddMosToTable();

	assert(InMosList());

	InstallSignalHandlers();

	safe_ostrstream	ss;
	ss << Name() << " start";
	gTraceLog->SendMessage(ss.str());
	while (!mTerminate)	// loop forever
		{
		try {
			_MessageHandler(mMessageQueue.Get());
			}
		catch(CError &err)
			{
			SendMessage(new CMosMessageType<nmr_string>
				(CMOS_SYS_INTERNAL_ERROR,new nmr_string(err.data())));
			}
		catch(...)
			{
			SendMessage(new CMosMessageType<nmr_string>
				(CMOS_SYS_INTERNAL_ERROR,new nmr_string("Unknown exception")));
			}

		}

	assert(InMosList());
	RemoveMosFromTable();
	assert(!InMosList());

	safe_ostrstream	s2;
	s2 << Name() << " stop";
	gTraceLog->SendMessage(s2.str());
}

//////////////////////////////////////////////////////////////////////////
// SendMessage
//
// Used to send a message to this mos. The message will be deleted by this 
// mos after it is handled. It is important that this method uses
// PutCanFail(). Since call backs tend to message themselves, if we
// use Put() then a potential deadlock can occur if the message queue
// gets full. The message cant be put because the queue is full, but
// the thread can't empty the queue because its trying to put.
//
// INPUT:
//	message - pointer to the message. It must be declared
//		on the heap as the MessageHandler will delete it
////////////////////////////////////////////////////////////////////////////
void
CMos::SendMessage(CMosMessage *message)
{
	mNoMoreMutex.Lock();
	if (mNoMoreMessages)
		{
		mNoMoreMutex.Unlock();
		delete message;
		return;
		}
	mNoMoreMutex.Unlock();

	if (!mMessageQueue.PutCanFail(message))
		{
		safe_ostrstream s;
		s << "Mos Queue Overflow in " << Name();
		gTraceLog->SendMessage(s.str());
		delete message;
		}
	// this is probably a good place to yield
	CCThread::Yield();
}

void
CMos::TerminateAll(void)
{
	stack_MutexLock sml(&sMosTableMutex);

	Mos_Table_t::reverse_iterator i;

	for (i = sMosTable.rbegin(); i != sMosTable.rend(); ++i)
			(*i).first->Terminate();		
}

//////////////////////////////////////////////////////////////////////////
// SendMessageIfSpaceElseBlock
//
// Used to send a message to this mos. The message will be deleted by this 
// mos after it is handled. This send will block if the queue is
// full beyond a given percentage
//
// INPUT:
//	message - pointer to the message. It must be declared
//		on the heap as the MessageHandler will delete it
//	percent	- beyond this value the call blocks until the queue empties
//
// OUTPUT:
//	returns true if it blocked
////////////////////////////////////////////////////////////////////////////
int				
CMos::SendMessageIfSpaceElseBlock(CMosMessage *message,int percent)
{
	mNoMoreMutex.Lock();
	if (mNoMoreMessages)
		{
		mNoMoreMutex.Unlock();
		delete message;
		return 0;
		}
	mNoMoreMutex.Unlock();

	int block = mMessageQueue.PutIfSpaceElseBlock(message,percent);

	// this is probably a good place to yield
	CCThread::Yield();
	return block;
}

int				
CMos::SendMessageIfSpaceElseFail(CMosMessage *message,int percent)
{
	mNoMoreMutex.Lock();
	if (mNoMoreMessages)
		{
		mNoMoreMutex.Unlock();
		delete message;
		return 0;
		}
	mNoMoreMutex.Unlock();

	int result = mMessageQueue.PutIfSpaceElseFail(message,percent);
	if (!result)
		delete message;

	// this is probably a good place to yield
	CCThread::Yield();
	return result;
}

////////////////////////////////////////////////////////////////////////////
// _MessageHandler
//
// protected handler method to allow automatic deletion of message. You 
// should override MessageHandler() to implement your handler
//
// INPUT:
//	message	 - the message to handle
//////////////////////////////////////////////////////////////////////////
void
CMos::_MessageHandler(CMosMessage *message)
{
	stack_Deleter<CMosMessage> sd(message);

	if (message->GetCode() == CMOS_SYS_TERMINATE)
		mTerminate = 1;
	else
		MessageHandler(message);
}

///////////////////////////////////////////////////////////////////////////////
// Terminate
//
// Sends a terminate message to thread
///////////////////////////////////////////////////////////////////////////////
void
CMos::Terminate(void)
{
	mNoMoreMutex.Lock();
	mNoMoreMessages = 1;
	mNoMoreMutex.Unlock();
	mMessageQueue.Put(new CMosMessage(CMOS_SYS_TERMINATE));

	CCThread::Yield();

}

///////////////////////////////////////////////////////////////////////////////
// MessageQueuePercentFull
//
// How full is the mos's message queue.
//
// OUTPUT:
//	percentage as an integer
///////////////////////////////////////////////////////////////////////////////
int
CMos::MessageQueuePercentFull(void) const
{
	return mMessageQueue.PercentFull();
}

int
CMos::NumObjectsInQueue(void) const
{
	return mMessageQueue.NumObjectsInQueue();
}


///////// SIGNAL HANDLING

// problem with linux pthread headers. Put prototype here
#ifdef OLDLINUX
extern "C" {void signal(int,void (*sig)(int));}
#endif

///////////////////////////////////////////////////////////////////////////////
// SignalCatch
//
// invoked from Unix signal mechanism to handle signals on a thread by
// thread basis.
//
// INPUT:
//	sig - the signal
///////////////////////////////////////////////////////////////////////////////
void 
CMos::SignalCatch(int sig)
{
#ifndef	WIN32

	signal(sig,SignalCatch);

	sSignal = sig;
	if (!sSigKill)
		sSigKill = (sig == SIGTERM);

#endif
}

///////////////////////////////////////////////////////////////////////////////
// InstallSignalHandlers
//
// Install Unix signal handlers for this object to allow rerouting as
// CMosMessages
///////////////////////////////////////////////////////////////////////////////
void
CMos::InstallSignalHandlers(void)
{
#if ((!defined WIN32) && (!defined __POWERPC__))

    signal(SIGPIPE, SignalCatch);
    signal(SIGHUP,  SignalCatch);
	signal(SIGTERM, SignalCatch);	
    signal(DUCKSIGNAL, SignalCatch);		

#endif
}

const char *
CMos::Name(void)
{
	return "CMos";
}

void
CMos::Dump(ostream &o,const char *indent)
{
	nmr_string	ndent(indent);
	ndent += "   ";

	o << indent << "CMos:" << endl
		<< indent << "   mTerminate: " << mTerminate << endl
		<< indent << "   mNoMoreMessages: " << mNoMoreMessages << endl
		<< indent << "   mNoMoreMutex: " << endl;
	mNoMoreMutex.Dump(o,ndent);
	o << indent << "   mMessageQueue: " << endl;
	mMessageQueue.Dump(o,ndent);
	CCThread::Dump(o,indent);
}



