/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CPacket.h>
#include	<CPacketSimple.h>
#include	<CPacketType.h>
#include	<CDataList.h>
#include	<CTCPSocket.h>
#include	<nmr_string.h>
#include	<stack_aids.h>

/*****************************************************************************
CPacket
	The base class for packets sent via CPortTransmitter and read by
	CPortMonitor. Packets read and write themselves to sockets. In order
	to accomplish this. Packets and their types must be registered with
	the packet subsystem. Some frequently used types are automatically
	registered. Others can be registered by calling RegisterPacketType().

	CPacket cannot be used on its own. You must use the dataless subclass
	CPacketSimple or the the data class template CPacketType
******************************************************************************/

int				CPacket::sPacketTableMax = -1;
int				CPacket::sPacketTableInitialized = 0;
CCMutex			CPacket::sPacketTableMutex;
packet_init_func_t*	
				CPacket::sPacketInitTable = NULL;

// This forces InitializeSystemPackets to get called before the
// application starts!!!!
class dummy_class {
public:
	dummy_class(void){ CPacket::InitializeSystemPackets();}
	~dummy_class(void){ CPacket::DestroySystemPackets();}
};

static dummy_class dummy;

//////////////////////// statics //////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// InitializeSystemPackets
//
// register commonly used packet types with the packet subsystem
///////////////////////////////////////////////////////////////////////////////
int
CPacket::InitializeSystemPackets(void)
{
	assert(!sPacketTableInitialized);
	RegisterPacketType(PACKET_TYPE_SIMPLE,
								CPacketSimple::_Creator);
	RegisterPacketType(PACKET_TYPE_I16,
								CPacketType<I16>::_Creator);
	RegisterPacketType(PACKET_TYPE_STRING,
								CPacketType<nmr_string>::_Creator);
	// no longer used, left in for compatibility
//	RegisterPacketType(PACKET_TYPE_CINETSPEC,
//								CPacketType<CInetSpec*>::_Creator);
	RegisterPacketType(PACKET_TYPE_I16LIST,
								CPacketType<CDataList<I16> >::_Creator);
	RegisterPacketType(PACKET_TYPE_I32LIST,
								CPacketType<CDataList<I32> >::_Creator);
	RegisterPacketType(PACKET_TYPE_I32,
								CPacketType<I32 >::_Creator);
	sPacketTableInitialized = 1;
	return 1;
}

void
CPacket::DestroySystemPackets(void)
{
	assert(sPacketTableInitialized);
	
	delete [] sPacketInitTable;
	sPacketInitTable = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// RegisterPacketType
//
// Register a packet type with the packet subsystem. 
//
// INPUT:
//	type_code			- the packet type code
//	packet_init_func_t	- the callback to create a packet of the given type
///////////////////////////////////////////////////////////////////////////////
void	
CPacket::RegisterPacketType(int type_code,packet_init_func_t init_func)
{
	stack_MutexLock sml(&sPacketTableMutex);

	if (sPacketTableMax < type_code)
		{
		packet_init_func_t	*temp = new packet_init_func_t[type_code+1];
		int	x;
		for (x = 0; x <= type_code; ++x)
			temp[x] = NULL;
		for (x = 0; x <= sPacketTableMax; ++x)
			temp[x] = sPacketInitTable[x];
		delete [] sPacketInitTable;
		sPacketInitTable = temp;
		sPacketTableMax = type_code;
		}
	sPacketInitTable[type_code] = init_func;
}

//////////////////////////////////////////////////////////////////////////////
// Receive
//
// Receive a packet from a socket.
//
// INPUT:
//	socket - the socket to read from
//
// OUTPUT:
//	p - the newly created packet
//
// throw exception if error
///////////////////////////////////////////////////////////////////////////////
CPacket *
CPacket::Receive(CTCPSocket &socket)
{
	int	code,packet_type,protocol_code;
	u_short	u[3];
	CPacket *p = NULL;

	try 
		{
		socket.Read((char *)u,sizeof(u));

		code 			= ntohs(u[0]);
		packet_type 	= ntohs(u[1]);
		protocol_code 	= ntohs(u[2]);
		
		if (sPacketTableMax < packet_type || packet_type < 0)
			throw PacketTypeException(
					"CPacket::Receive packet type out of range",packet_type);
	
		if (!sPacketInitTable[packet_type])
			throw PacketTypeException(
					"CPacket::No creator for packet type",packet_type);
		p =(*sPacketInitTable[packet_type])(code,packet_type,protocol_code);
		p -> _Read(socket);
		}
	catch(...)
		{
		delete p;
		throw;
		}

	return p;
}

///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Send
//
// Send a packet via a socket
//
// INPUT:
//	socket - the socket to write to
//
///////////////////////////////////////////////////////////////////////////////
void
CPacket::Send(CTCPSocket &socket)
{
	u_short	u[3];

	// set up write vector
	u[0] 	= htons(mCode);
	u[1] 	= htons(mType);
	u[2] 	= htons(mProtocolCode);

	socket.Write((char *)u,sizeof(u));
	_Write(socket);
}

//////////////////////////////////////////////////////////////////////////////
// CPacket
//
// Constructor.
//
// INPUT:
//	code			- the message code of the packet
//	packet_type		- the packet type code
//	protocol_code	- the protocol associated with this packet
///////////////////////////////////////////////////////////////////////////////
CPacket::CPacket(int code,int packet_type,int protocol_code)
{
	assert(sPacketTableInitialized);
	mCode = code;
	mType = packet_type;
	mProtocolCode = protocol_code; 
}

///////////////////////////////////////////////////////////////////////////////
// ~CPacket
//
// Destructor
///////////////////////////////////////////////////////////////////////////////
CPacket::~CPacket(void)
{
}

///////////////////////////////////////////////////////////////////////////////
// GetProtocol
//
// Return the protocol code associated with this packet
//
// OUTPUT:
//	the protocol code
///////////////////////////////////////////////////////////////////////////////
int
CPacket::GetProtocol(void)
{
	return mProtocolCode;
}

///////////////////////////////////////////////////////////////////////////////
// GetType
//
// Return the type of this packet
//
// OUTPUT:
//	the type code
///////////////////////////////////////////////////////////////////////////////
int
CPacket::GetType(void)
{
	return mType;
}

///////////////////////////////////////////////////////////////////////////////
// GetCode
//
// Return the message code of this packet. This code will be duplicated as
// the code for the CMosMessage that may encapsulate the packet later on
//
// OUTPUT:
//	The message code
///////////////////////////////////////////////////////////////////////////////
int
CPacket::GetCode(void)
{
	return mCode;
}

///////////////////////////////////////////////////////////////////////////////
// DetachData
//
// Do nothing used in subclasses
///////////////////////////////////////////////////////////////////////////////
void
CPacket::DetachData(void)
{
	// do nothing, used in subclasses
}

size_t
CPacket::CalculateSize(void)
{
	return (sizeof(u_short) * 3);
}

const char *
CPacket::Name(void)
{
	return "CPacket";
}

void
CPacket::Dump(ostream &o,const char *indent)
{
	o << indent << "CPacket:" << endl;
	o << indent << "   mCode: " << mCode << endl;
	o << indent << "   mType: " << mType << endl;
	o << indent << "   mProtocolCode: " << mProtocolCode << endl;
}

ostream& operator<<(ostream &o,CPacket &p)
{
	p.Dump(o,"");
	return o;
}

ostream& operator<<(ostream &o,CPacket *p)
{
	if (p)
		p->Dump(o,"");
	else
		o << " NULL " << endl;
	return o;
}

