/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CPortMonitor.h>
#include	<CProtocol.h>
#include	<CMosMessageType.h>
#include	<CPacketType.h>
#include	<CTCPSocket.h>
#include	<nmr_string.h>
#include	<stack_aids.h>

/******************************************************************************
CPortMonitor
	used by one or more CProtocol's to receive packets. The protocol's must
	register themselves with the CPortMonitor via RegisterProtocol().
	When the CPortMonitor gets a packet it uses the protocol code to
	look up the registered CProtocol's and dispatch the packet. The
	packet is dispatched inside a CMosMessageType<CPacket> whose message
	code is that of the packet.

	If an error occurs, all registered protocols are informed via a
	CMosMessage and the CPorMonitor terminates
******************************************************************************/

#define PM_OK		(CMOS_SYS_INIT_OK)
#define	PM_FETCH	(CMOS_SYS_FIRST_USER_EVENT)

///////////////////////////////////////////////////////////////////////////////
// CPortMonitor
//
// Constructor
//
// INPUT:
//	queue_size	- size of message queue
//	socket		- socket to read from
///////////////////////////////////////////////////////////////////////////////
CPortMonitor::CPortMonitor(int queue_size,CTCPSocket *socket)
	: CMos(queue_size)
{
	assert(socket);
	mDispatchTableMax = -1;
	mDispatchTable = NULL;
	mSocket = socket;
	mErrorInducedIgnore = 0;
	mIgnoreErrors = 0;
}

void
CPortMonitor::SetIgnoreErrors(int i)
{
	mIgnoreErrors = i;
}

///////////////////////////////////////////////////////////////////////////////
// ~CPortMonitor
//
// Destructor
///////////////////////////////////////////////////////////////////////////////
CPortMonitor::~CPortMonitor(void)
{
	stack_MutexLock sml(&mDispatchTableMutex);

	delete [] mDispatchTable;
	mDispatchTable = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// RegisterProtocol
//
// Register a protocol module with the portmonitor for dispatch. 
//
// INPUT:
//	protocol - the protocol module to register
//
// OUTPUT:
//	1	- all is well
//	0	- protocol is null or already exists in table
///////////////////////////////////////////////////////////////////////////////
int
CPortMonitor::RegisterProtocol(CProtocol *protocol)
{
	stack_MutexLock sml(&mDispatchTableMutex);

	int	result,code;

	if (protocol == NULL) return 0;

	code = protocol->GetProtocolCode();

	if (mDispatchTableMax < code)
		{
		Dispatch_Table_t temp = new protocol_map_entry[code+1];
		int	x;
		for (x = 0; x <= code; ++x)
			temp[x] = NULL;
		for (x = 0; x <= mDispatchTableMax; ++x)
			temp[x] = mDispatchTable[x];

		delete [] mDispatchTable;
		mDispatchTable = temp;
		mDispatchTableMax = code;
		}

	result =  (mDispatchTable[code] == NULL) ;
	if (result)
		mDispatchTable[code] = protocol;

	return result;
}


///////////////////////////////////////////////////////////////////////////////
// Fetch
//
// Send a message to the port monitor to get another packet
///////////////////////////////////////////////////////////////////////////////
void
CPortMonitor::Fetch(void)
{
	SendMessage(new CMosMessage(PM_FETCH));
}

///////////////////////////////////////////////////////////////////////////////
// MessageHandler
//
// virtual handler which monitors the socket and creates message packets.
// these packets are then sent to the appropriate protocol. If the
// protocol does not exist then an internal error is sent to all
// threads in dispatch table. If an io error occurs the
// same happens. Either error will then terminate this thread
//
// INPUT:
//		message - the message
///////////////////////////////////////////////////////////////////////////////
void
CPortMonitor::MessageHandler(CMosMessage *message)
{
	CPacket	 *p = NULL;

	switch(message->GetCode())
		{
		case PM_OK:
			break; 
		case PM_FETCH:
			if (!mErrorInducedIgnore)
				{
				try {
					p = CPacket::Receive(*mSocket);
					int	code = p->GetProtocol();
					if (code > mDispatchTableMax || (!mDispatchTable[code]))
						{
						if (!mIgnoreErrors)
							SendToAllInDispatch(CMOS_SYS_INTERNAL_ERROR);
						}
					else
						{
						CProtocol *protocol = mDispatchTable[code];
						protocol->MessageMe(new CMosMessageType<CPacket>(
							p->GetCode(),p));
						}
					}
				catch(...)
					{
					// send io error to all threads including myself
					delete p;
					mErrorInducedIgnore = 1;
					if (!mIgnoreErrors)
						SendToAllInDispatch(CMOS_SYS_IO_ERROR);
					}
				}
			break;
		default:
			// if we don't know what it is we assume an error
			if (!mIgnoreErrors)
				SendToAllInDispatch(CMOS_SYS_INTERNAL_ERROR);
			break;
		}
}

void
CPortMonitor::Terminate(void)
{
	mIgnoreErrors = 1;
	mErrorInducedIgnore = 1;
	CMos::Terminate();

	if (mSocket)
		mSocket->BreakBlockingIO(this);
}

///////////////////////////////////////////////////////////////////////////////
// SendToAllInDispatch
//
// Send a message to every protocol in the dispatch table
//
// INPUT:
//	the message to send
///////////////////////////////////////////////////////////////////////////////
void
CPortMonitor::SendToAllInDispatch(int code)
{
	for (int x = 0; x <= mDispatchTableMax; ++x)
		{
		CProtocol	*p = mDispatchTable[x];
		if (p)
			p -> MessageMe(new CMosMessage(code));
		}
}

const char *
CPortMonitor::Name(void)
{
	return "CPortMonitor";
}

void
CPortMonitor::Dump(ostream &o,const char *indent)
{
	nmr_string	nmr(indent);
	nmr += "   ";

	o << indent << "CPortMonitor: " << endl;
	o << indent << "   mDispatchTableMutex: " << endl;
	mDispatchTableMutex.Dump(o,nmr);
	o << indent << "   mSocket: ";
	if (mSocket)
		{
		o << endl;
		mSocket->Dump(o,nmr);
		}
	else
		{
		o << " NULL " << endl;
		}

	o << indent << "   mDispatchTableMax: " << mDispatchTableMax << endl;
	o << indent << "   mErrorInducedIgnore: " << mErrorInducedIgnore << endl;
	o << indent << "   mIgnoreErrors: " << mIgnoreErrors << endl;
	o << indent << "   mDispatchTable: " << endl;
	nmr += "   ";
	for (int x = 0; x <= mDispatchTableMax; ++x)
		{
		if (mDispatchTable[x]) 
			mDispatchTable[x]->Dump(o,nmr);
		else
			o << nmr << " NULL " << endl;
		}
	CMos::Dump(o,nmr);
}



