/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CPortTransmitter.h>
#include	<CTCPSocket.h>
#include	<CMosMessageType.h>
#include	<CPacket.h>
#include	<CProtocol.h>
#include	<nmr_string.h>


/******************************************************************************
CPortTransmitter
	Class that sends packets from a CProtocol. If an error occurs it 
	notifies the CProtocol via a CMosMessage and terminates.
******************************************************************************/

///////////////////////////////////////////////////////////////////////////////
// CPortTransmitter
//
// Constructor
//
// INPUT:
//	queue_size	- size of message queue
//	socket		- socket to use for transmitting
//////////////////////////////////////////////////////////////////////////////
CPortTransmitter::CPortTransmitter(int queue_size,CTCPSocket *socket)
	: CMos(queue_size)
{
	mSocket = socket;
	mErrorInducedIgnore = 0;
	mIgnoreErrors = 0;
}

///////////////////////////////////////////////////////////////////////////////
// ~CPortTransmitter
//
// Destructor
///////////////////////////////////////////////////////////////////////////////
CPortTransmitter::~CPortTransmitter(void)
{
}

void
CPortTransmitter::SetIgnoreErrors(int i)
{
	mIgnoreErrors = i;
}

void
CPortTransmitter::Terminate(void)
{
	mIgnoreErrors = 1;
	mErrorInducedIgnore = 1;
	CMos::Terminate();

	if (mSocket)
		mSocket->BreakBlockingIO(this);
}

#define	CPT_INIT	CMOS_SYS_INIT_OK
#define	CPT_SEND	CMOS_SYS_FIRST_USER_EVENT
#define CPT_IOERR	CMOS_SYS_IO_ERROR

///////////////////////////////////////////////////////////////////////////////
// SendPacket
//
// CProtocol objects call this to send a packet
//
// INPUT:
//	who 	- the sending protocol
//	p		- the packet to send
///////////////////////////////////////////////////////////////////////////////
void
CPortTransmitter::SendPacket(CProtocol *who,CPacket *p)
{
	CPTData	*data = new CPTData(who,p);
	CMosMessageType<CPTData> *message = 
		new CMosMessageType<CPTData>(CPT_SEND,data);
	SendMessage(message);
}

///////////////////////////////////////////////////////////////////////////////
// SendPacketIfSpaceElseBlock
//
// CProtocol objects call this to send a packet. This call uses space
// thresholding to limit queue flooding
//
// INPUT:
//	who 	- the sending protocol
//	p		- the packet to send
// percent	- limit beyond which call will block
//
// OUTPUT:
//	return true if it blocked
///////////////////////////////////////////////////////////////////////////////
int
CPortTransmitter::SendPacketIfSpaceElseBlock(CProtocol *who,CPacket *p,
																int percent)
{
	CPTData	*data = new CPTData(who,p);
	CMosMessageType<CPTData> *message = 
		new CMosMessageType<CPTData>(CPT_SEND,data);
	return SendMessageIfSpaceElseBlock(message,percent);
}

///////////////////////////////////////////////////////////////////////////////
// SendPacketImmediately
//
// Bypass usual messaging system and send the packet out right now.
// Use with extreme caution since this bypasses the normal safeguards.
// You could easily get an ioerror if two threads try to write at
// the same time.
// The packet is deleted as it would be if a normal send was done.
//
// INPUT:
//	who		- sending protocol
//	p		- packet to send
///////////////////////////////////////////////////////////////////////////////
void
CPortTransmitter::SendPacketImmediately(CProtocol *who,CPacket *p)
{
	try {
		p->Send(*mSocket);
		}
	catch(...)
		{
		who->MessageMe(new CMosMessage(CMOS_SYS_IO_ERROR));
		}
	delete p;
}

///////////////////////////////////////////////////////////////////////////////
// RawSendPacket
//
// Do the packet send
//
// INPUT:
//	p	- the packet
//
///////////////////////////////////////////////////////////////////////////////
void
CPortTransmitter::RawSendPacket(CPacket *p)
{
	p->Send(*mSocket);
}

///////////////////////////////////////////////////////////////////////////////
// MessageHandler
//
// Actually sends the packets queued up by SendPacket(). If an error 
// occurs than the CProtocol that queued the request is notified via
// a CMosMessage.
//
// INPUT:
//	message - message to process
///////////////////////////////////////////////////////////////////////////////
void
CPortTransmitter::MessageHandler(CMosMessage *message)
{
	switch(message->GetCode())
		{
		case CPT_INIT:
			break;
		case CPT_SEND:
			{
			CMosMessageType<CPTData>	*p;
			if (!mErrorInducedIgnore)
				{
				try 
					{
					p = (CMosMessageType<CPTData>*)message;

					RawSendPacket ((p->GetData())->mPacket);
					}
				catch(...)
					{
					// send io error to calling thread
					mErrorInducedIgnore = 1;
					if (!mIgnoreErrors)
						((p->GetData())->mWho)->MessageMe(
							new CMosMessage(CMOS_SYS_IO_ERROR));
					}
				}
			}
			break;
		default:
			break;
		}
}

void
CPortTransmitter::Dump(ostream &o,const char *indent)
{
	nmr_string	nmr(indent);
	nmr += "   ";

	o << indent << "CPortTransmitter: " << endl;
	o << indent << "   mErrorInducedIgnore: " << mErrorInducedIgnore << endl;
	o << indent << "   mIgnoreErrors: " << mIgnoreErrors << endl;
	o << indent << "   mSocket: ";
	if (mSocket)
		{
		o << endl;
		mSocket->Dump(o,nmr);
		}
	else
		{
		o << " NULL " << endl;
		}

	CMos::Dump(o,nmr);
}

const char *
CPortTransmitter::Name(void)
{
	return "CPortTransmitter";
}

////////////data type of private CPortTransmitter messages//////////////////
CPTData::CPTData(CProtocol *who,CPacket *p)
{
	assert(who);
	assert(p);
	mWho = who;
	mPacket = p;
}

CPTData::~CPTData(void)
{
	assert(mPacket);
	delete mPacket;
}	

void
CPTData::Dump(ostream &o,const char *indent)
{
	nmr_string	nmr(indent);
	nmr += "   ";
	o << indent << "CPTData: " << endl;
	o << indent << "   mWho: " << endl;
	if (mWho)
		mWho->Dump(o,nmr);
	o << indent << "   mPacket: " << endl;
	if (mPacket)
		mPacket->Dump(o,nmr);

}


ostream& operator<<(ostream &o,CPTData &c)
{
	c.Dump(o,"");
	return o;
}

ostream& operator<<(ostream &o,CPTData *c)
{
	if (c)
		c->Dump(o,"");
	else 
		o << " NULL " << endl;
	return o;
}

