/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CSocket.h>
#include	<CCThread.h>
#include	<stack_aids.h>
/*
	Abstract socket class. Use either CTCPSocket or CUDPSocket
*/

#ifdef	WIN32
class OpenWinsock
{
	public:
		OpenWinsock(void)
			{
			WSADATA	junk;
			int 	result;
			result = WSAStartup(MAKEWORD(1,1),&junk);
			if (result != 0)
				{
				cout << "WSAStartup error" << endl << flush;
				exit(1);
				}
			}
		~OpenWinsock(void)
			{
			WSACleanup();
			}
};

OpenWinsock ow;
#endif

const unsigned long	CSocket::AnyAddress = INADDR_ANY;
CCMutex				CSocket::sGetHostByNameMutex;

CSocket::CSocket()
{
	mSocketClosed = 1;
}

///////////////////////////////////////////////////////////////////////////////
//
// Create a socket from an already open descriptor
//
///////////////////////////////////////////////////////////////////////////////
CSocket::CSocket(SOCKET socket_descriptor)
{
	mSocketDescriptor = socket_descriptor;
	mSocketClosed = 0;
	mLastBoundPort = 0;
}

CSocket::~CSocket(void)
{
	if (!mSocketClosed)
		Close();
}

void
CSocket::Close(void)
{
	if (mSocketClosed) return;
	int result = SOCKET_close(mSocketDescriptor);
#ifdef	WIN32
	if (result)
#else
	if (result < 0)
#endif
		throw CloseException("CSocket::Close",FETCH_SOCKET_ERROR);
	mSocketClosed = 1;
}

///////////////////////////////////////////////////////////////////////////////
// BreakBlockingIO
//
// Call this to break thread t out of a blocking I/O.
//
///////////////////////////////////////////////////////////////////////////////
void
CSocket::BreakBlockingIO(CCThread *t)
{
#if ((defined WIN32) || (defined __POWERPC__))
	// under windows and mac our only choice is to close the socket
	try {Close();} catch(...) {}
#else
	// under unix we send a signal to the thread. Closing the socket
	// can cause most unixes to hang
	t->Kill(DUCKSIGNAL);
#endif
}

I16
CSocket::GetLastBoundPort(void)
{
	return ntohs(mLastBoundPort);
}

void
CSocket::Bind(unsigned long addr,int port)
{
	struct sockaddr_in serv_addr;
	int	result;

	memset(&serv_addr,0,sizeof(serv_addr));
	serv_addr.sin_family 		= AF_INET;
	serv_addr.sin_addr.s_addr 	= htonl(addr);
	serv_addr.sin_port 			= htons(port);
	result = SOCKET_bind(mSocketDescriptor,(struct sockaddr *)&serv_addr,
															sizeof(serv_addr));
#ifdef	WIN32
	if (result)
#else
	if (result < 0)
#endif
		throw BindException("CSocket::Bind",FETCH_SOCKET_ERROR);

	int len = sizeof(serv_addr);
	SOCKET_getsockname(mSocketDescriptor,(struct sockaddr *)&serv_addr,&len);
	mLastBoundPort = serv_addr.sin_port;
}

///////////////////////////////////////////////////////////////////////////////
// SetSockOpt
//
// Encapsulation of unix setsockopt (see unix docs)
//
///////////////////////////////////////////////////////////////////////////////
void	
CSocket::SetSockOpt(int level,int optname,const void *optval,int optlen)
{
	int	result;

	// WIN32 needs char* cast
	result = SOCKET_setsockopt(mSocketDescriptor,level,optname,(char *)optval,
																	optlen);
#ifdef	WIN32
	if (result)
#else
	if (result < 0)
#endif
		throw SetSockOptException("CSocket::SetSockOpt",FETCH_SOCKET_ERROR);
}

///////////////////////////////////////////////////////////////////////////////
// GetSockOpt
//
// Encapsulation of unix getsockopt (see unix docs)
//
///////////////////////////////////////////////////////////////////////////////
void	
CSocket::GetSockOpt(int level,int optname,void *optval,int *optlen)
{
	int	result;

	// WIN32 needs char* cast
	result = SOCKET_getsockopt(mSocketDescriptor,level,optname,(char *)optval,
																	optlen);
#ifdef	WIN32
	if (result)
#else
	if (result < 0)
#endif
		throw GetSockOptException("CSocket::GetSockOpt",FETCH_SOCKET_ERROR);
}

void
CSocket::SetOption_ReuseAddr(bool b)
{
	int one = (b ? 1 : 0);
	SetSockOpt(SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(one));
}

void			
CSocket::SetOption_ReceiveBuffer(int siz)
{
	int	bufsize = siz;
	SetSockOpt(SOL_SOCKET,SO_RCVBUF,&bufsize,sizeof(bufsize));
}

void			
CSocket::SetOption_SendBuffer(int siz)
{
	int	bufsize = siz;
	SetSockOpt(SOL_SOCKET,SO_SNDBUF,&bufsize,sizeof(bufsize));
}

unsigned long
CSocket::GetHostByName(const char *host)
{
	// UNDER LINUX THREADS gethostbyname LEAKS!!!!!
	stack_MutexLock sml(&sGetHostByNameMutex);

	struct hostent *he = SOCKET_gethostbyname(host);
	if (he)
		return *((unsigned long *)he->h_addr);
	else
		return 0;
}

void
CSocket::GetSocketInfo(unsigned long &addr,int &port)
{
	struct sockaddr_in serv_addr;
	int	len;

	memset(&serv_addr,0,sizeof(serv_addr));
	len 	= sizeof(serv_addr);
	int result = SOCKET_getsockname(mSocketDescriptor,
										(struct sockaddr *)&serv_addr,&len);
#ifdef	WIN32
	if (result)
#else
	if (result < 0)
#endif
		throw SocketInfoException("CSocket::GetSocketInfo",FETCH_SOCKET_ERROR);

	addr	= ntohl(serv_addr.sin_addr.s_addr);
	port 	= ntohs(serv_addr.sin_port);
}


const char *
CSocket::Name(void)
{
	return "CSocket";
}

void
CSocket::Dump(ostream &o,const char *indent)
{
	o << indent << "CSocket: " << endl
		<< indent << "   mSocketClosed: " << mSocketClosed << endl
		<< indent << "   mSocketDescriptor: " << mSocketDescriptor << endl;
}
