/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include	<CTcpServer.h>
#include	<CTcpService.h>
#include	<CLogger.h>
#include	<CTCPSocket.h>
#include	<stack_aids.h>
#include	<safe_ostrstream.h>

/******************************************************************************
CTcpServer
	monitors a Tcp socket and hands off connections to CTcpServices.
******************************************************************************/

MD(CTcpServer,	Open,		CMOS_SYS_OPEN);
MD(CTcpServer,	Init,		CMOS_SYS_INIT_OK);
MD(CTcpServer,	IO_Error,	CMOS_SYS_IO_ERROR);
MD(CTcpServer,	INT_Error,	CMOS_SYS_INTERNAL_ERROR);

/*----------------------- Callbacks -----------------------------------------*/

///////////////////////////////////////////////////////////////////////////////
// CB_Init
//
// Called when CTcpServer starts up. You must override BuildTcpServices
// to construct and install the services for this server
///////////////////////////////////////////////////////////////////////////////
void
CTcpServer::CB_Init(CMosMessage *mess)
{
	TraceCallback("CB_Init");

	BuildTcpServices();

	SEND_MESS_Open();
}

///////////////////////////////////////////////////////////////////////////////
// CB_Error
//
// Called if an error occurs. Logs the error and terminates
///////////////////////////////////////////////////////////////////////////////
void
CTcpServer::CB_Error(CMosMessage *mess)
{
	TraceCallback("CB_Error");

	safe_ostrstream buf;
	buf << this << "::CB_Error " << mess;
	gErrorLog->SendMessage(buf.str());

	Terminate();
}

///////////////////////////////////////////////////////////////////////////////
// CB_Open
//
// Opens the socket and waits for connections
///////////////////////////////////////////////////////////////////////////////
void
CTcpServer::CB_Open(CMosMessage *mess)
{
	TraceCallback("CB_Open");

	assert(!mSocket);

try{
	mSocket = new CTCPSocket;
	SetSocketOptions(mSocket);
	mSocket->Bind(CSocket::AnyAddress,mPort);
	mSocket->Listen(mPortBacklog);

	while (mAcceptingConnections)
		{
		SOCKET	s;
		s = mSocket->Accept();
			{
			// use a block here in case an exception is thrown	
			stack_MutexLock sml(&mAcceptingMutex);
			
			if (mAcceptingConnections)
				HandleConnection(s);
			}
		}
	}
catch(...)
	{
	if (!CMos::sSigKill && !mIgnoreAcceptError) // clean shutdown on terminate
		throw;
	}
}



#define	STATE_INITIAL	0
#define	STATE_READY		1
#define	STATE_ACTIVE	2


CFsm::state_table_entry_t	CTcpServer::sTable[] = 
{
{STATE_INITIAL,	MESS_Init,		STATE_READY,	FCT(CTcpServer::CB_Init)},
{STATE_INITIAL,	MESS_INT_Error,	STATE_INITIAL,	FCT(CTcpServer::CB_Error)},
{STATE_INITIAL,	MESS_IO_Error,	STATE_INITIAL,	FCT(CTcpServer::CB_Error)},

{STATE_READY,	MESS_Open,		STATE_READY,	FCT(CTcpServer::CB_Open)},
{STATE_READY,	MESS_INT_Error,	STATE_INITIAL,	FCT(CTcpServer::CB_Error)},
{STATE_READY,	MESS_IO_Error,	STATE_INITIAL,	FCT(CTcpServer::CB_Error)},
};

///////////////////////////////////////////////////////////////////////////////
// CTcpServer
//
// Constructor. Attaches the state table.
//
// INPUT:
//	queue_size		- size of message queue
//	port			- port to listen on
//	port_backlog	- number of connections that can be queued up without
//							being handled.
///////////////////////////////////////////////////////////////////////////////
CTcpServer::CTcpServer(int queue_size,int max_services,int port,int port_backlog)
	: CFsm(queue_size)
{
	AttachStateTable(sTable,sizeof(sTable) / sizeof(state_table_entry_t));

	mIgnoreAcceptError = 0;
	mAllServicesTableMaxSize= mServiceTableMaxSize 	= max_services;
	mAllServicesTableSize 	= mServiceTableSize 	= 0;
	mServices 		= new service_table_t[max_services];
	mAllServices 	= new service_table_t[max_services];

	for (int x = 0; x < max_services; ++x)
		{
		mServices[x] 	= NULL;
		mAllServices[x] = NULL;
		}

	mSocket 		= NULL;
	mPort 			= port;
	mPortBacklog 	= port_backlog;
	mAcceptingConnections = 1;
}

///////////////////////////////////////////////////////////////////////////////
// ~CTcpServer
//
// Destructor. Deletes the socket if it's still around
///////////////////////////////////////////////////////////////////////////////
CTcpServer::~CTcpServer(void)
{
	for (int x = 0; x < mAllServicesTableMaxSize; ++x)
		{
		assert(mAllServices[x]);
		mAllServices[x]->Service_Join();
		delete mAllServices[x];
		}

	delete mSocket;
	mSocket = NULL;
	delete mServices;
	mServices = NULL;
	delete mAllServices;
	mAllServices = NULL;
}

void
CTcpServer::Terminate(void)
{
	// force a break out of any blocked IO
	mAcceptingMutex.Lock();
	mAcceptingConnections = 0;
	mAcceptingMutex.Unlock();

	// we must post a terminate to ourselves first, otherwise if
	// out service threads post errors, we'll keep recalling
	// CTcpServer::Terminate()
	CFsm::Terminate();

	// now shut down the services. This must be done before
	// will kill ourselves because Solaris only allows one
	// outstanding signal

	for (int x = 0; x < mAllServicesTableMaxSize; ++x)
		{
		assert(mAllServices[x]);
		mAllServices[x]->Service_Terminate();
		}

	mIgnoreAcceptError = 1;
	if (mSocket)
		mSocket->BreakBlockingIO(this);
}

void	
CTcpServer::BuildTcpServices(void)
{
	for (int x = 0; x < mServiceTableMaxSize; x++)
		{
		CTcpService	*service = CreateTcpService(x);
		service->Service_Start();
		AddTcpService(service);
		RegisterForShutdownTcpService(service);
		}
}

///////////////////////////////////////////////////////////////////////////////
// AddTcpService
//
// Adds a CTcpService to the available queue
//
// INPUT:
//	service - the service to add to the queue
///////////////////////////////////////////////////////////////////////////////
void
CTcpServer::AddTcpService(CTcpService *service)
{
	stack_MutexLock sml(&mServiceMutex);

	assert(mServiceTableSize < mServiceTableMaxSize);
	mServices[mServiceTableSize++] = service;
}

void
CTcpServer::RegisterForShutdownTcpService(CTcpService *service)
{
	assert(mAllServicesTableSize < mAllServicesTableMaxSize);

	mAllServices[mAllServicesTableSize++] = service;
}

///////////////////////////////////////////////////////////////////////////////
// RemoveTcpService
//
// Removes a CTcpService from the queue so it can be used
//
// OUTPUT:
//	the CTcpService that has been removed. Returns NULL if none are
//	available.
///////////////////////////////////////////////////////////////////////////////
CTcpService*
CTcpServer::RemoveTcpService(void)
{
	stack_MutexLock sml(&mServiceMutex);

	return (mServiceTableSize > 0 ? mServices[--mServiceTableSize] : 
		(CTcpService *)NULL);
}

///////////////////////////////////////////////////////////////////////////////
// HandleConnection
//
// Decide what to do with the new connection. The default is to
// get a service and hand off the connection
//
// INPUT:
//	fd - the file descriptor for the new connection
///////////////////////////////////////////////////////////////////////////////
void
CTcpServer::HandleConnection(SOCKET fd)
{
	CTcpService	*service;
	service = RemoveTcpService();
	if (service != NULL)
		{
		service->SEND_MESS_Open(new SOCKET(fd));
		}
	else
		{
		gLogLog->SendMessage("TcpServer - no services available");
		SOCKET_close(fd);
		}
}

const char *
CTcpServer::Name(void)
{
	return "CTcpServer";
}

void
CTcpServer::Dump(ostream &o,const char *indent)
{
	stack_MutexLock sml(&mServiceMutex);

	nmr_string	nmr(indent);
	nmr += "   ";

	o << indent << Name() << ":" << endl;
	o << indent << "   mIgnoreAcceptError: " << mIgnoreAcceptError << endl;
	o << indent << "   mServiceTableSize: " << mServiceTableSize << endl;
	o << indent << "   mServiceTableMaxSize: " << mServiceTableMaxSize << endl;
	o << indent << "   mServices: " << endl;
	for (int x = 0; x < mServiceTableSize; ++x)
		if (mServices[x]) mServices[x]->Dump(o,nmr);

	o << indent << "   mPortBacklog: " << mPortBacklog << endl;
	o << indent << "   mPort: " << mPort << endl;

	o << indent << "   mSocket: " << endl;
	if (mSocket) mSocket->Dump(o,nmr);
	o << indent << "   mServiceMutex: " << endl;
	mServiceMutex.Dump(o,nmr);
	CFsm::Dump(o,nmr);
}


///////////////////////////////////////////////////////////////////////////////
// 	void BuildTcpServices(void);
//
// Abstract. You override this to construct and install the services
// for this server.
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// 	void SetSocketOptions(CTCPSocket *);
//
// Abstract. Called immediately after the socket is created. Override
// to set the socket options you want.
///////////////////////////////////////////////////////////////////////////////







