/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef ffirChunk_included
#define ffirChunk_included

// generic riff chunk parser...
//
// Alan S. Rojer
// The Duck Corp.
// 2 Sept 1997

#include "ffirFCC.h"
class ivaIO;

class ffirChunkHeader {
public:
	~ffirChunkHeader() {}
	ffirChunkHeader() {}
	ffirChunkHeader(ffirFCC FCC, ulong size) : _FCC(FCC), _size(size) {}
	const ffirFCC& FCC() const {return _FCC;}
	ulong size() const {return _size;}
	static int read(ivaIO&, ffirChunkHeader&);
	virtual int write(ivaIO&) const;
private:
	ffirFCC _FCC;
	ulong _size;
};

inline
int operator==(const ffirChunkHeader &h1, const ffirChunkHeader &h2)
{
	return h1.FCC() == h2.FCC() && h1.size() == h2.size();
}

inline std::ostream& operator<<(std::ostream &os, const ffirChunkHeader &h) {
	os << h.FCC() << " " << h.size();
	return os;
}

class ffirListChunkHeader : public ffirChunkHeader {
	// ...yes, we're inheriting from a class with data, which is generally
	// a bad idea, so we have to be extra careful here.
public:
	~ffirListChunkHeader() {}
	ffirListChunkHeader(ulong size, ffirFCC list_FCC)
	: ffirChunkHeader(ffirFCC::LIST(), size), _list_FCC(list_FCC) {}
		// ...WARNING: reversing the fcc and ulong arguments will not
		// give a compiler error, since fcc/ulong are silently cross-convertible.
		// But it's a bug for sure to reverse them, so be careful! I was bitten.
	ffirListChunkHeader(ulong size)
	: ffirChunkHeader(ffirFCC::RIFF(), size), _list_FCC(ffirFCC::AVI()) {}
	ffirFCC list_FCC() const {return _list_FCC;}
	int write(ivaIO&) const;
private:
	ffirFCC _list_FCC;
};

inline std::ostream& operator<<(std::ostream &os, ffirListChunkHeader &h) {
	os << h.FCC() << "(" << h.list_FCC() << ") " << h.size();
	return os;
}

// The Handler below plays the role of traditional callback functions
// in C.  It's an abstract base classes; you inherit from them and 
// put whatever 'client data' you need into the private parts of
// your derived class.  This gets you type-safe callbacks with user
// user data.  No more void* casts!

// It's perfectly safe to do fread()s in the handler; the parser
// saves and restores file position around the callbacks.

class ffirHandler {
public:
	virtual ~ffirHandler() {}
	virtual int handle_chunk(ivaIO &, const ffirChunkHeader&, uint level)=0;
		// ...file position is immediately after chunk header (at start of data)
	virtual int handle_list_chunk(ivaIO &, const ffirChunkHeader&, ffirFCC, 
		uint level, int &skip)=0;
		// ...file position is immediately after list 4cc in data part of chunk
		// ...(4 bytes into the data part of the chunk)
		// ...handler may set 'skip' non-zero to skip the current chunk 
		// ...(ie, don't parse its subchunks)
};

class ffirParser {
public:
	// factory...
	~ffirParser() {}
	ffirParser(ivaIO &f, ffirHandler *h=0) : _f(f), _handler(h) {}
	void set_handler(ffirHandler *h) {_handler = h;} // ...better to do it at	
		// at construction time!
	int parse() const;
private:
	// members...
	ivaIO &_f;
	ffirHandler *_handler;
private:
	// methods...
	int parse(const ffirChunkHeader &, uint level) const;
	int parse_list(const ffirChunkHeader &, uint level) const;
	int fail(const char *method, const char *message) const;
};

#endif // ffirChunk_included
