/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef intelFil_h
#define intelFil_h 1

// Read/write multibyte primitive types from/to intel-format character strings.
// Timothy S. Murphy 7/24/95.


/* Character strings are in intel x86 format, that is:

   Chars are 8 bits, shorts are 16 bits, longs are 32 bits,
   and everything is stored from low-order to high-order.

   This is what you want for .avi, .wav, and other microsoft-defined files
   but may not be what you want elsewhere (e.g., mac-defined files).

   Of course, we make no assumptions about the machine we're running on;
   in fact, that's the entire point of this package.

   Be careful if your architecture has different sizes than these.

   Note that these sizes of chars, shorts, and longs are exactly
   the acceptable minima defined by ANSI C/C++ and that any program
   using these types _portably_ may safely archive them in a byte
   stream using the methods given here.

   BTW, an int is not guaranteed to be any bigger than a short. */


#include "iva_timstyle.h"
#include "iva_machine.h"


// Convenient stream syntax, advances supplied pointer as we go.

// This stuff could be made faster in the canonical case.


class intelReader {
	const uchar *s;
 public:
	intelReader( const char *src =0): s( (const uchar *)src) {}
	intelReader( const uchar *src): s( src) {}
	
	operator const uchar *() const { return s;}

	inline friend bool operator==( const intelReader& r1, const intelReader& r2)
		{ return r1.s == r2.s;}
	inline friend bool operator<( const intelReader& r1, const intelReader& r2)
		{ return r1.s < r2.s;}

	intelReader& operator+=( int i) { s += i;  return *this;}
	intelReader& operator-=( int i) { s -= i;  return *this;}

	inline friend intelReader& operator>>( intelReader& r, char& c)
		{ c = *(char *)r.s++;  return r;}
	inline friend intelReader& operator>>( intelReader& r, schar& c)
		{ c = *(schar *)r.s++;  return r;}
	inline friend intelReader& operator>>( intelReader& r, uchar& c)
		{ c = *r.s++;  return r;}

	inline friend intelReader& operator>>( intelReader& r, short& i) {
#		if targetCPUisLittleEndian  &&  USHRT_MAX == 65535
			i = *(short *)r.s;
#		else
			i = r.s[0] + (r.s[1]<<8);
#		endif
		r.s += 2;
		return r;
	}
	inline friend intelReader& operator>>( intelReader& r, ushort& i) {
#		if targetCPUisLittleEndian  &&  USHRT_MAX == 65535
			i = *(ushort *)r.s;
#		else
			i = r.s[0] + (r.s[1]<<8);
#		endif
		r.s += 2;
		return r;
	}
	inline friend intelReader& operator>>( intelReader& r, long& L) {
#		if targetCPUisLittleEndian  &&  ULONG_MAX == 4294967295
			L = *(long *)r.s;
#		else
			L = r.s[0] + (r.s[1]<<8) + (r.s[2]<<16) + (r.s[3]<<24);
#		endif
		r.s += 4;
		return r;
	}
	inline friend intelReader& operator>>( intelReader& r, ulong& L) {
#		if targetCPUisLittleEndian  &&  ULONG_MAX == 4294967295
			L = *(ulong *)r.s;
#		else
			L = r.s[0] + (r.s[1]<<8) + (r.s[2]<<16) + (r.s[3]<<24);
#		endif
		r.s += 4;
		return r;
	}

};


#ifdef WIN32 // busted VC compiler
class intelWriter;
intelWriter& operator<<( intelWriter& w, uchar c);
intelWriter& operator<<( intelWriter& w, char c);
intelWriter& operator<<( intelWriter& w, schar c);
intelWriter& operator<<( intelWriter& w, short i);
intelWriter& operator<<( intelWriter& w, ushort i);
intelWriter& operator<<( intelWriter& w, long L);
intelWriter& operator<<( intelWriter& w, ulong L);
#endif

class intelWriter {
	uchar *s;
 public:
	intelWriter( uchar *src =0): s( src) {}

	operator uchar *() const { return s;}

	inline friend bool operator==( const intelWriter& r1, const intelWriter& r2)
		{ return r1.s == r2.s;}
	inline friend bool operator<( const intelWriter& r1, const intelWriter& r2)
		{ return r1.s < r2.s;}

	inline friend intelWriter& operator<<( intelWriter& w, uchar c)
		{ *w.s++ = c;  return w;}
	inline friend intelWriter& operator<<( intelWriter& w, char c)
		{ *w.s++ = (uchar)c;  return w;}
	inline friend intelWriter& operator<<( intelWriter& w, schar c)
		{ *w.s++ = (uchar)c;  return w;}

	inline friend intelWriter& operator<<( intelWriter& w, short i) {
#		if targetCPUisLittleEndian  &&  USHRT_MAX == 65535
			*(short *)w.s = i;  w.s += 2;
#		else
			*w.s++ = (uchar)i;  *w.s++ = (uchar)(i>>8);
#		endif
		return w;
	}
	inline friend intelWriter& operator<<( intelWriter& w, ushort i) {
#		if targetCPUisLittleEndian  &&  USHRT_MAX == 65535
			*(ushort *)w.s = i;  w.s += 2;
#		else
			*w.s++ = (uchar)i;  *w.s++ = (uchar)(i>>8);
#		endif
		return w;
	}
	inline friend intelWriter& operator<<( intelWriter& w, long L) {
#		if targetCPUisLittleEndian  &&  ULONG_MAX == 4294967295
			*(long *)w.s = L;  w.s += 4;
#		else
			*w.s++ = (uchar)L;  *w.s++ = (uchar)(L>>=8);
			*w.s++ = (uchar)(L>>=8);  *w.s++ = (uchar)(L>>8);
#		endif
		return w;
	}
	inline friend intelWriter& operator<<( intelWriter& w, ulong L) {
#		if targetCPUisLittleEndian  &&  ULONG_MAX == 4294967295
			*(ulong *)w.s = L;  w.s += 4;
#		else
			*w.s++ = (uchar)L;  *w.s++ = (uchar)(L>>=8);
			*w.s++ = (uchar)(L>>=8);  *w.s++ = (uchar)(L>>8);
#		endif
		return w;
	}
};

#endif	// intelFil_h
