/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//	$Id: iva_Parser.cxx,v 1.1 2001/12/12 18:47:50 nradisch Exp $
//
// Alan S. Rojer
// The Duck Corp.
// Mon Sep  8 08:26:34 EDT 1997

#include "iva_Parser.h"
#include "ivaIO.h"
#include <vector>
using namespace std;

typedef iva_Parser Parser;
typedef iva_Handler Handler;

iva_Parser::iva_Parser(ivaIO &f, Handler *avi_handler) 
: _avi_handler(avi_handler), _chunk_parser(f)
{
	_chunk_parser.set_handler(this);
}

iva_Parser::~iva_Parser()
{
}

int
iva_Parser::fail(const char *method, const char *message) const
{
	cerr << "iva_Parser::" << method << ": " << message << endl;
	return 0;
}

int
iva_Parser::parse() const
{
	return _chunk_parser.parse();
}

int
iva_Parser::handle_chunk(ivaIO &f, const ffirChunkHeader &h, uint)
{
	if (h.FCC() == ffirFCC::avih()) return handle_avih(f, h);
	if (h.FCC() == ffirFCC::idx1()) return handle_idx1(f, h);
	return 1;
}

int
iva_Parser::handle_avih(ivaIO &f, const ffirChunkHeader &h)
{
	const char *me = "handle_avih";
	uchar *buffer = new uchar [h.size()]; // ...should be owner
	if (!buffer) return fail(me, "malloc failed");
	_last_header_mark = f.Getpos();
	if (f.Read(buffer, h.size()) != h.size()) return fail(me, "read failed");
	ivaHeader *p_header = new ivaHeader;
	if (!p_header) return fail(me, "malloc failed");
	if (!ivaHeader::read_from_buffer(buffer, p_header))
		return fail(me, "couldn't translate avih from buffer");
	_avi_handler->accept_header(p_header);
	delete [] buffer;
	return 1;
}

typedef pmsio_AVIINDEXENTRY IndexEntry;
#ifdef NEEDS_STL_ALLOCATOR
typedef vector<IndexEntry, allocator<IndexEntry> > IndexEntryVector_t;
#else
typedef vector<IndexEntry> IndexEntryVector_t;
#endif

int
iva_Parser::handle_idx1(ivaIO &f, const ffirChunkHeader &h)
{
	const char *me = "handle_idx1";
	IndexEntryVector_t index;
	long index_offset;

	{ // scarf the index into an array...
		uchar *buffer = new uchar [h.size()]; // ...should be owner
		if (!buffer) return fail(me, "malloc failed");
		if (f.Read(buffer, h.size()) != h.size()) return fail(me, "read failed");
		int i, n = h.size()/IndexEntry::size();
		uchar *u = buffer;
		for (i=0; i<n; i++, u += IndexEntry::size()) {
			IndexEntry entry;
			if (!IndexEntry::read_from_buffer(u, &entry)) 
				return fail(me, "couldn't read index entry");
			index.push_back(entry);
		}
		delete [] buffer;
	}

	if (index.size() == 0) return 1;

	{ // determine the addressing mode for the index...
		// first scan for the first data block...
		// (we allow for the possibility that the index might reorder the file)
		IndexEntryVector_t::iterator ii = index.begin();
		IndexEntryVector_t::iterator first = ii;
		//const IndexEntry *first = ii;
		++ii; // ...we know there's at least one
		while (ii != index.end()) {
			if ((*ii).dwChunkOffset < (*first).dwChunkOffset) 
				first = ii;
			++ii;
		}
		long chunk_offset = (*first).dwChunkOffset;
		// now examine the chunk offset...
		if (chunk_offset == _movi_offset) {
			// simplest case: absolute indexing...
			index_offset = 0;
		}
		else if (chunk_offset == 0) {
			// addressing from movi_offset...
			index_offset = _movi_offset;
		}
		else if (chunk_offset == 4) {
			// addressing from start of movi data chunk...
			index_offset = _movi_offset -4;
		}
		else if (chunk_offset == 12) {
			// addressing from start of movi chunk...
			index_offset = _movi_offset - 12;
		}
		else {
			char s[1024];
			sprintf(s,"unknown indexing mode: chunk offset %lX, movi offset %lX",
				chunk_offset, _movi_offset);
			return fail(me, s);
		}
		// NOTE: why not just assume index_offset = movi_offset - chunk_offset
		// if the chunk offset is nonzero??? -- ask Eric and John.
	}

	{ // give the client her notifications...
		IndexEntryVector_t::iterator ii = index.begin();
		while (ii != index.end()) {
			(*ii).dwChunkOffset += index_offset;
			if (!_avi_handler->accept_index_entry(*ii)) 
				return fail(me, "index entry handler failed");
			++ii;
		}
	}
	return 1;
}

int
iva_Parser::handle_list_chunk(ivaIO &f, const ffirChunkHeader &h,
	ffirFCC list_FCC, uint, int &skip)
{
	if (list_FCC == ffirFCC::movi()) {
		skip = 1;
		return handle_movi(f); 
	}
	if (list_FCC == ffirFCC::strl()) {
		skip = 1; 
		return handle_strl(f, h); 
	}
	return 1;
}

int
iva_Parser::handle_strl(ivaIO &f, const ffirChunkHeader &h)
{
	const char *me = "handle_strl";

	// here's what we're assembling...
	ivaStreamHeader *p_stream_header = 0;
	ivaVideoHeader *p_video_stream_header = 0;
	ivaAudioHeader *p_audio_stream_header = 0;
	ivaFrame *p_other_header = 0;
	ivaDriverData *p_driver_data = 0;

	size_t bytes_to_read = h.size() - 4; // ...list 4cc code was already read

	// read the strh dude...
	ffirChunkHeader strh_header;
	if (!ffirChunkHeader::read(f, strh_header)) 
		return fail(me, "strh header read failed");
	uchar *strh_buffer = new uchar [strh_header.size()];
	if (!strh_buffer) return fail(me, "strh malloc failed");
	_last_header_mark = f.Getpos();
	if (f.Read(strh_buffer, strh_header.size()) != strh_header.size()) 
		return fail(me, "strh read failed");
	bytes_to_read -= strh_header.size()  + 8; // ...including header
	p_stream_header = new ivaStreamHeader;
	if (!ivaStreamHeader::read_from_buffer(strh_buffer, p_stream_header))
		return fail(me, "couldn't translate strh from buffer");

	// read the appropriate strf...
	ffirChunkHeader strf_header;
	if (!ffirChunkHeader::read(f, strf_header)) 
		return fail(me, "strf header read failed");
	uchar *strf_buffer = new uchar [strf_header.size()];
	if (!strf_buffer) return fail(me, "strf malloc failed");
	_last_format_mark = f.Getpos();
	if (f.Read(strf_buffer, strf_header.size()) != strf_header.size()) 
		return fail(me, "strf read failed");
	bytes_to_read -= strf_header.size()  + 8; // ...including header
	if ((ulong)ffirFCC::vids() == p_stream_header->fccType) {
		p_video_stream_header = new ivaVideoHeader;
		if (!ivaVideoHeader::read_from_buffer(strf_buffer, 
			p_video_stream_header))
			return fail(me, "couldn't translate strf/vids from buffer");
	}
	else if ((ulong)ffirFCC::auds() == p_stream_header->fccType) {
		p_audio_stream_header = new ivaAudioHeader;
		if (!ivaAudioHeader::read_from_buffer(strf_buffer, 
			p_audio_stream_header))
			return fail(me, "couldn't translate strf/auds from buffer");
	}
	else {
		return fail(me, "unexpected fcctype in strh");
		// to do... handle "other"
	}

#if(0)
	if (bytes_to_read > 0) {
		// read the strd dude, if any...
		ffirChunkHeader strd_header;
		if (!ffirChunkHeader::read(f, strd_header))
			return fail(me, "strd header read failed");
		uchar *strd_buffer = strd_buffer = new uchar [strd_header.size()];
		if (!strd_buffer) return fail(me, "strd malloc failed");
		if (fread(strd_buffer, strd_header.size(), 1, f) != 1) 
			return fail(me, "strd read failed");
		p_driver_data = new ivaOwnerFrame(strd_buffer, strd_header.size());
		if (!p_driver_data) return fail(me, "malloc failed");
		bytes_to_read -= strd_header.size()  + 8; // ...including header
	}
#endif

	if ((ulong)ffirFCC::vids() == p_stream_header->fccType) {
		_avi_handler->accept_video_header(p_stream_header, 
			p_video_stream_header, p_driver_data);
		{
		// fixes leak 7/6/98
		delete [] strh_buffer;
		delete [] strf_buffer;
		return 1;
		}
	}
	else if ((ulong)ffirFCC::auds() == p_stream_header->fccType) {
		_avi_handler->accept_audio_header(p_stream_header, 
			p_audio_stream_header, p_driver_data);
	}
	else {
		return fail(me, "unexpected fcctype in strh");
		// to do... handle "other"
	}

	// buffers should be 'owned'...this code leaks on error...
	delete [] strh_buffer;
	delete [] strf_buffer;
	return 1;
}

int
iva_Parser::handle_movi(ivaIO &f)
{
	const char *me = "iva_Parser::handle_movi";
	_movi_offset = f.Getpos();
	_avi_handler->accept_movi_offset(_movi_offset);
	return 1;
}

