/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "iva_Reader.h"
#include "ffirChunk.h"
#include "iva_defines.h"
#include "ivaIO.h"
#include <assert.h>

#define FAIL(who, what, why) \
	{cerr << who << ": " << what << ": " << why << endl; return 0;}

iva_Reader::iva_Reader(ivaIO &io)
: _f(io),  _fname(0), _header(0), _movi_offset(0), _index_found(0)
{
}

iva_Reader::~iva_Reader()
{
	close();
	ivaReadStreamBase_t::iterator i = _streams.begin();
	while (i != _streams.end()) 
		{
		delete (*i);
		++i;
		}
	delete _header;
	delete [] _fname;
}

bool
iva_Reader::open(const char *fname)
{
	const char *me = "iva_Reader::open";
	delete [] _fname;

	_fname = new char[strlen(fname) + 1];
	if (!_fname) FAIL(me, "_fname", "malloc failed")
	strcpy(_fname, fname);

	// open the file...
	if (!_f.Open(fname,ivaIO::READ)) FAIL(me, _fname, "couldn't open")

	// define handler and parse the file...
	iva_Parser parser(_f, this);
	if (!parser.parse()) FAIL(me, _fname, "parse failed")

	return true;
}

int
iva_Reader::is_open(void) const
{
	return _f.IsOpen();
}


int
iva_Reader::close(int recalc)
{
	if (_f.IsOpen())
		_f.Close();
	delete [] _fname;
	_fname = 0;
	return true;
}

const ivaAudioReadStream *
iva_Reader::audio_stream(uint i) const
{
	if (i >= _audio_streams.size()) return 0;
	return _audio_streams[i];
}

const ivaVideoReadStream *
iva_Reader::video_stream(uint i) const
{
	if (i >= _video_streams.size()) return 0;
	return _video_streams[i];
}

const ivaOtherReadStream *
iva_Reader::other_stream(uint i) const
{
	if (i >= _other_streams.size()) return 0;
	return _other_streams[i];
}


// handlers...
int
iva_Reader::accept_audio_header(ivaStreamHeader *stream_header,
							ivaAudioHeader *audio_header,
							ivaDriverData *driver_data)
{
	iva_AudioReadStream *rs = 
		new iva_AudioReadStream(this, stream_header,audio_header, driver_data);
	_streams.push_back(rs); 
	_stream_bases.push_back(rs->base()); 
	_audio_streams.push_back(rs);
	return 1;
}

int
iva_Reader::accept_video_header(ivaStreamHeader *stream_header,
	ivaVideoHeader *video_header,
	ivaDriverData *driver_data)
{
	iva_VideoReadStream *rs = new iva_VideoReadStream(this, stream_header,
		video_header, driver_data);
	_streams.push_back(rs); 
	_stream_bases.push_back(rs->base()); 
	_video_streams.push_back(rs);
	return 1;
}

int
iva_Reader::accept_other_header(ivaStreamHeader *stream_header,
	ivaFrame *other_header,
	ivaDriverData *driver_data)
{
	iva_OtherReadStream *rs = new iva_OtherReadStream(this, stream_header,
		other_header, driver_data);
	_streams.push_back(rs); 
	_stream_bases.push_back(rs->base()); 
	_other_streams.push_back(rs);
	return 1;
}

int
iva_Reader::accept_index_entry(const ivaIndexEntry &entry)
{
	const char *me = "iva_Reader::accept_index_entry";
	ffirFCC FCC(entry.ckid);
	ulong n;
	if (!FCC.extract_stream_number(n)) return 1;
	// verify n...
	if (n < _streams.size()) 
		_stream_bases[n]->_index.push_back(entry);
	return 1;
}

int
iva_Reader::read_frame(ivaFrame &frame, const ivaIndexEntry &ie) const
{
	const char *me = "iva_Reader::read_frame";
	ulong offset = ie.dwChunkOffset;
	ulong length = ie.dwChunkLength;
	if (length > frame.buffer_size()) FAIL(me, "frame error", "buffer too small")
	if (_f.Seek(offset,ivaIO::SET) == ivaIO::sSeekOffsetError) FAIL(me, _fname, "seek failed")
	ffirChunkHeader h;
	if (!ffirChunkHeader::read(_f, h))  FAIL(me, _fname, "header read failed")
#if(0)
	{ // verify?
		if (h.FCC() == chunk_FCC && h.size() == aligned_size) return 1;
		cerr << "reader: " << me 
				<< ": expected " << chunk_FCC << "; got " << h.FCC() << endl;
		cerr << "reader: " << me 
				<< ": expected " << aligned_size << "; got " << h.size() << endl;
		return 0;
	}
#endif
//cerr << me << ": reading frame at offset " 
//<< hex << offset << " of length " << length << endl;
	if (length > 0) {
		if (_f.Read(frame.buffer(), length) != length) 
			FAIL(me, _fname, "fread failed")
		frame.length() = length;
	}
	else frame.length() = 0;
	return 1;
}

iva_StreamReader::iva_StreamReader(
	const iva_ReadStreamBase *stream, 
	ulong start, ulong stop)
: _stream(stream), _current(start), _stop(stop)
{
	if (_stop > _stream->_index.size()) 
		_stop = _stream->_index.size();
}

int
iva_StreamReader::read(ivaFrame &frame, int &is_key, ffirTCC &twocc)
{
	if (_current >= _stop) return 0;
	ivaIndexEntry ie = _stream->_index[_current ++];
	is_key = (ie.dwFlags & AVIIF_KEYFRAME);
	//ffirTCC tcc(ie.ckid);
	twocc = ffirTCC(ie.ckid);
	return _stream->_reader->read_frame(frame, ie);
}

int	
iva_StreamReader::read(U32 index,ivaFrame &frame,int &is_key, ffirTCC &twocc)
{
	if (index >= _stream->_index.size()) return 0;
	ivaIndexEntry ie = _stream->_index[index];
	is_key = (ie.dwFlags & AVIIF_KEYFRAME);
	//ffirTCC tcc(ie.ckid);
	twocc = ffirTCC(ie.ckid);
	return _stream->_reader->read_frame(frame, ie);
}


ivaStreamReader *
iva_ReadStreamBase::stream_reader(ulong start, ulong stop) const
{
	return new iva_StreamReader(this, start, stop);
}

U32		
iva_ReadStreamBase::total_data(void) const
{
	U32 total = 0;

	IvaIndexEntryVector_t::const_iterator i;
	for (i = _index.begin(); i != _index.end(); ++i)
		total += (*i).dwChunkLength;

	return total;
}

U32		
iva_ReadStreamBase::indicies(void) const
{
	return _index.size();
}


U32		
iva_ReadStreamBase::chunk_size(U32 index) const
{
	assert(index <= _index.size());
	return (_index[index].dwChunkLength);
}


