/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef iva_Reader_included
#define iva_Reader_included

// $Id: iva_Reader.h,v 1.1 2001/12/12 18:47:52 nradisch Exp $
// Reference implementation of avi file reader service.
//
// Alan S. Rojer
// The Duck Corp.
// 1 Sept. 1997

#include "iva.h"
#include "ffirFCC.h"
#include "iva_Parser.h"

#include <vector>
using namespace std;

class ivaIO;

typedef pmsio_AVIINDEXENTRY iva_IndexEntry;

class iva_Reader;

class iva_ReadStreamBase : public ivaReadStreamBase {
private:
	// ordinary members...
	const iva_Reader *_reader;
	ivaStreamHeader *_header;
	ivaDriverData *_driver_data;
private:
	// index management...
#ifdef NEEDS_STL_ALLOCATOR
	typedef vector<ivaIndexEntry, allocator<ivaIndexEntry> > IvaIndexEntryVector_t;
#else
	typedef vector<ivaIndexEntry> IvaIndexEntryVector_t;
#endif
	IvaIndexEntryVector_t _index;
	friend class iva_Reader;
	friend class iva_StreamReader;
public:
	// factory...
	iva_ReadStreamBase(
		const iva_Reader *reader, 
		ivaStreamHeader *header, 
		ivaDriverData *driver_data)
	: _reader(reader), _header(header), _driver_data(driver_data) {}
	~iva_ReadStreamBase() {
		delete _header;
		delete _driver_data;
	}
public:
	// ivaReadStreamBase methods...
	const ivaReader *reader() const;
	const ivaStreamHeader *header() const {return _header;}
	const ivaDriverData *driver_data() const {return _driver_data;}
	ivaStreamReader *stream_reader(ulong start=0, ulong end = -1) const;
	U32		indicies(void) const;
	U32		total_data(void) const;
	U32		chunk_size(U32 index) const;
};

template <class StreamFormat>
class iva_ReadStream : public ivaReadStream<StreamFormat> {
private:
	// ordinary members...
	iva_ReadStreamBase _base;
	StreamFormat *_stream_format;
public:
	// factory...
	~iva_ReadStream() {
		delete _stream_format;
	}
	iva_ReadStream(const iva_Reader *r, ivaStreamHeader *sh,
		StreamFormat *sf, ivaDriverData *dd)
	: _base(r, sh, dd), _stream_format(sf) {}
public:
	// ivaReadStream methods...
	const ivaReader *reader() const {return _base.reader();}
	const ivaStreamHeader *header() const {return _base.header();}
	const StreamFormat *stream_format() const {return _stream_format;}
	const ivaDriverData *driver_data() const {return _base.driver_data();}
	ivaStreamReader *stream_reader(ulong start=0, ulong end = -1) const {
		return _base.stream_reader(start, end);
	}
public:
	iva_ReadStreamBase *base() {return &_base;}
	U32		indicies(void) const { return _base.indicies(); }
	U32		total_data(void) const { return _base.total_data(); }
	U32		chunk_size(U32 index) const { return _base.chunk_size(index); }
};

typedef iva_ReadStream<ivaVideoHeader> iva_VideoReadStream;
typedef iva_ReadStream<ivaAudioHeader> iva_AudioReadStream;
typedef iva_ReadStream<ivaOtherHeader> iva_OtherReadStream;

class iva_Reader : public ivaReader, private iva_Handler {
private:
	// member variables...
	ivaIO &_f;
	char *_fname;
	ivaHeader *_header;
#ifdef NEEDS_STL_ALLOCATOR
	typedef vector <iva_AudioReadStream*,allocator<iva_AudioReadStream*> > iva_AudioReadStream_t;
	typedef vector <iva_VideoReadStream*,allocator<iva_VideoReadStream*> > iva_VideoReadStream_t;
	typedef vector <iva_OtherReadStream*,allocator<iva_OtherReadStream*> > iva_OtherReadStream_t;
	typedef vector <ivaReadStreamBase*,allocator<ivaReadStreamBase*> > ivaReadStreamBase_t; // ...for deletion
	typedef vector <iva_ReadStreamBase*,allocator<iva_ReadStreamBase*> > iva_ReadStreamBase_t; // ...for access to index
#else
	typedef vector <iva_AudioReadStream*> iva_AudioReadStream_t;
	typedef vector <iva_VideoReadStream*> iva_VideoReadStream_t;
	typedef vector <iva_OtherReadStream*> iva_OtherReadStream_t;
	typedef vector <ivaReadStreamBase*> ivaReadStreamBase_t; // ...for deletion
	typedef vector <iva_ReadStreamBase*> iva_ReadStreamBase_t; // ...for access to index
#endif

	iva_AudioReadStream_t _audio_streams;
	iva_VideoReadStream_t _video_streams;
	iva_OtherReadStream_t _other_streams;
	ivaReadStreamBase_t _streams; // ...for deletion
	iva_ReadStreamBase_t _stream_bases; // ...for access to index
	long _movi_offset;
	int _index_found;
private:
	// iva_Handler methods...
	int accept_header(ivaHeader *header) {_header = header; return 1;}
		// ...leaks on reuse!}
	int accept_audio_header(ivaStreamHeader*, ivaAudioHeader*, ivaDriverData*);
	int accept_video_header(ivaStreamHeader*, ivaVideoHeader*, ivaDriverData*);
	int accept_other_header(ivaStreamHeader*, ivaOtherHeader*, ivaDriverData*);
	int accept_movi_offset(long movi_offset) {
		_movi_offset = movi_offset;
		return 1;
	}
	int accept_index_entry(const ivaIndexEntry&);
public:
	// factory methods...
	iva_Reader(ivaIO &io);
	~iva_Reader();
public:
	// ivaFile methods...
	bool open(const char *fname);
	int is_open() const;
	int close(int recalc = 0);
	const ivaHeader *header() const {return _header;}
public:
	// ivaReader methods...
	const ivaVideoReadStream *video_stream(uint i) const ;
	const ivaAudioReadStream *audio_stream(uint i) const ;
	const ivaOtherReadStream *other_stream(uint i) const ;
public:
	// the real workhorse; used by StreamReader...
	int read_frame(ivaFrame &, const ivaIndexEntry&) const;
};

inline const ivaReader *
iva_ReadStreamBase::reader() const {
	return _reader;
}

class iva_StreamReader : public ivaStreamReader {
private:
	const iva_ReadStreamBase *_stream;
	ulong _current, _stop;
public:
	// factory:
	iva_StreamReader(const iva_ReadStreamBase *stream, ulong start, ulong stop);
	~iva_StreamReader() {}
public:
	// ivaStreamReader methods...
	const ivaReadStreamBase *stream() const {return _stream;}
	int read(ivaFrame &, int &is_key, ffirTCC &twocc);
	int	read(U32 index,ivaFrame &,int &is_key, ffirTCC &twocc);
};

#endif // iva_Reader_included
