/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "iva_Writer.h"
#include "ffirChunk.h"
#include "iva_defines.h"
#include "ivaIO.h"
#include "danbug.h"

#define FAIL(who, what, why) \
	{cerr << who << ": " << what << ": " << why << endl; return false;}


iva_WriteStreamBase::iva_WriteStreamBase(
	iva_Writer *writer,
	int stream_number,
	const ivaStreamHeader *header, 
	const ivaDriverData *driver_data)
: _writer(writer),
	_stream_number(stream_number),
	_header(new ivaStreamHeader(*header)),
	_driver_data(driver_data ? driver_data->clone() : 0),
	_frames_written(0),
	_bytes_written(0),
	_max_frame(0)
{
}

iva_WriteStreamBase::~iva_WriteStreamBase()
{
	delete _header;
	delete _driver_data;
}

int
iva_WriteStreamBase::write(const ivaFrame &frame, int is_key, ffirTCC twocc)
{
 dprintf("iva_WriteStreamBase::write");
	ffirFCC FCC(_stream_number, twocc);
	int result =  _writer->put_frame(FCC, frame, is_key);
	if (result) 
		{
		++ _frames_written;
		const size_t fl = frame.length();
		_bytes_written += fl;
		if (fl > _max_frame) _max_frame = fl;
		}
 dprintf(" exit iva_WriteStreamBase::write");
	return result;
}

ivaFrame *
iva_WriteStreamBase::write_getframe(size_t siz)
{
	return _writer->get_frame(siz);
}

int	
iva_WriteStreamBase::write_returnframe(ivaFrame *f,int is_key, ffirTCC twocc)
{
	ffirFCC FCC(_stream_number, twocc);
	int result = _writer->return_frame(FCC, f, is_key);
	if (result) 
		{
		++ _frames_written;
		const size_t fl = f->length();
		_bytes_written += fl;
		if (fl > _max_frame) _max_frame = fl;
		}
	return result;
}


void
iva_VideoWriteStream::header_fixup(void)
{
	_header->dwLength = _frames_written;
}

void
iva_AudioWriteStream::header_fixup(void)
{
	_header->dwLength = _bytes_written / _stream_format->nBlockAlign;
	_header->dwSampleSize = _stream_format->nBlockAlign;
}

void
iva_OtherWriteStream::header_fixup(void)
{
	_header->dwLength = _frames_written;
}

template <class StreamFormat>
iva_WriteStream<StreamFormat>::iva_WriteStream(
	iva_Writer *writer,
	int stream_number,
	const ivaStreamHeader *header,
	const StreamFormat *stream_format,
	const ivaDriverData *driver_data)
: iva_WriteStreamBase(writer, stream_number, header, driver_data),
	_stream_format(stream_format->clone())
{
}

template <class StreamFormat>
int
iva_WriteStream<StreamFormat>::write_stream_format(ivaIO &f)
{
	const char *me = "iva_WriteStream::write_stream_format";
	size_t length = _stream_format->length();
	ffirChunkHeader strf_header(ffirFCC::strf(), length);
	if (!strf_header.write(f)) FAIL(me, "strf header", "write failed")
	if (!_stream_format->write(f)) FAIL(me, "stream format", "write failed")
	// ensure word alignment...
	if (length % 2) 
	if (f.Seek(1,ivaIO::CUR) == ivaIO::sSeekOffsetError) FAIL(me, "word alignment", "seek failed")
	return 1;
}

/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////

ivaWriteStream *
iva_Writer::create_stream(
	const ivaStreamHeader *sh,
	const ivaVideoHeader *vh,
	const ivaDriverData *dd)
{
	int n = _streams.size();
	iva_VideoWriteStream *ws = new iva_VideoWriteStream(this, n, sh, vh, dd);
	_video_streams.push_back(ws);
	_streams.push_back(ws);
	return ws;
}

ivaWriteStream *
iva_Writer::create_stream(
	const ivaStreamHeader *sh,
	const ivaAudioHeader *vh,
	const ivaDriverData *dd)
{
	int n = _streams.size();
 dprintf("iva_Writer::create_stream");
	iva_AudioWriteStream *ws = new iva_AudioWriteStream(this, n, sh, vh, dd);
	_audio_streams.push_back(ws);
	_streams.push_back(ws);
	return ws;
}

ivaWriteStream *
iva_Writer::create_stream(
	const ivaStreamHeader *sh,
	const ivaOtherHeader *vh,
	const ivaDriverData *dd)
{
	int n = _streams.size();
	iva_OtherWriteStream *ws = new iva_OtherWriteStream(this, n, sh, vh, dd);
	_other_streams.push_back(ws);
	_streams.push_back(ws);
	return ws;
}


ivaFrame *
iva_Writer::get_frame(size_t siz)
{
	return new ivaOwnerFrame(siz);
}

int		
iva_Writer::return_frame(ffirFCC fcc, ivaFrame *f,int is_key)
{
	int result = put_frame(fcc,*f,is_key);
	delete f;
	return result;
}

int
iva_Writer::put_frame(ffirFCC FCC, const ivaFrame &frame, int is_key)
{
	const char *me = "iva_Writer::put_frame";
	size_t data_size = frame.length();
	ffirChunkHeader h(FCC, data_size);
	if (!h.write(_f)) FAIL(me, "chunk header", "write failed")

	if (!write_word_aligned(frame.buffer(), data_size))
		FAIL(me, "frame", "write failed")

	unsigned long position = _movi_bytes;
	unsigned long chunk_size = 8 + data_size; // ...including chunk header
	_movi_bytes += chunk_size;

	// pad with junk...
	if (_chunk_padding && (chunk_size % _chunk_padding)) {
		long padding_required = _chunk_padding - chunk_size % _chunk_padding;
		if (padding_required < 8) padding_required += _chunk_padding;
			// ...padding must be at least 8 byte due to header!
		long junk_data_size = padding_required - 8; // ...header again!
		ffirChunkHeader jh(ffirFCC::JUNK(), junk_data_size);
		if (!jh.write(_f)) FAIL(me, "junk header", "write failed");
		if (_f.Seek(junk_data_size,ivaIO::CUR) == ivaIO::sSeekOffsetError) 
			FAIL(me, "alignment", "seek failed")
		_movi_bytes += padding_required;
	}

	// update the index...
	iva_IndexEntry ie;
	ie.ckid = FCC;
	ie.dwFlags = (is_key ? AVIIF_KEYFRAME : 0);
	ie.dwChunkOffset = position;
	ie.dwChunkLength = frame.length();
	_index.push_back(ie);
	_idx1_bytes += ie.size();

	return 1;
}

bool
iva_Writer::open(const char *fname)
{
	const char *me = "iva_Writer::open";
	delete [] _fname;
 dprintf("iva_Writer::open");
	_fname = new char[strlen(fname) + 1];
	if (!_fname) FAIL(me, "_fname", "malloc failed")
	strcpy(_fname, fname);

	_movi_bytes = 4; // ...the 'movi' list header
	_idx1_bytes = 0;

	// open the file...
	if (!_f.Open(fname,ivaIO::WRITE)) FAIL(me, fname, "couldn't open");
	if (_f.Seek(_header_padding,ivaIO::SET) == ivaIO::sSeekOffsetError) 
		FAIL(me, "header", "seek failed")

	return true;
}

int 
iva_Writer::is_open() const
{
	return _f.IsOpen();
}

#if(0)
static void
showpos(ivaIO &f)
{
	long pos = f.Getpos();
	char s[32];
	sprintf(s, "%08lX\t", pos);
	cerr << s;
}
#endif

void
iva_Writer::recalc_framerate(void)
{
	if (_video_streams.size() < 1 || _audio_streams.size() < 1) return;
	iva_AudioWriteStream	*aud_stream = _audio_streams[0];
	iva_VideoWriteStream	*vid_stream = _video_streams[0];
	ivaStreamHeader *aud_info	= aud_stream->header();
	ivaStreamHeader *vid_info	= vid_stream->header();
	ivaVideoHeader	*vid_format = vid_stream->stream_format();
	ivaAudioHeader	*aud_format = aud_stream->stream_format();

//	vid_info->dwRate = vid_info->dwLength * 
//				aud_format->nChannels * aud_format->nSamplesPerSec * aud_format->wBitsPerSample;
//	vid_info->dwScale = aud_info->dwLength * 8;

	float rate = ((float)vid_info->dwLength) * 
				((float)aud_format->nChannels) * 
				((float)aud_format->nSamplesPerSec) * 
				((float)aud_format->wBitsPerSample);

///	dbm -- this code needs to change to deal with compressed audio:
	float scale = ((float)aud_info->dwLength) * 8 * aud_format->nBlockAlign;

// avoid divide by zero
	if (scale == 0)
		{
		rate = 30;
		scale = 1;
		}
	vid_info->dwRate = (int)((rate / scale) * 1000.0);
	vid_info->dwScale = 1000;
}

int
iva_Writer::close(int recalc_fr)
{
	const char *me = "iva_Writer::close";
	//if (!is_open()) return 1;

	// adjust stream info
	iva_WriteStreamBase_t::iterator i = _streams.begin();
	while (i != _streams.end()) {
		(*i)->header_fixup();
		++i;
	}
	if (recalc_fr)
		recalc_framerate();

	// write the index...
	ffirChunkHeader idx1_header(ffirFCC::idx1(), _idx1_bytes);
	if (!idx1_header.write(_f)) FAIL(me, "idx1 header", "write failed"); 
	iva_IndexEntry_t::const_iterator ii = _index.begin();
	while (ii != _index.end()) {
		if (!(*ii).write(_f)) FAIL(me, "index", "write failed");
		++ii;
	}

	// write the header...
	if (_f.Seek(0,ivaIO::SET) == ivaIO::sSeekOffsetError) FAIL(me, "header", "seek failed")

	// riff chunk header...
	unsigned long riff_bytes = (_header_padding - 8) // ...don't count riff header
		+ (_movi_bytes - 4)  // ...4 _movi bytes (list 4cc) are in the header
		+ (_idx1_bytes + 8); // ...index bytes + header
	ffirListChunkHeader riff_header(riff_bytes);
	if (!riff_header.write(_f)) FAIL(me, "riff header", "write failed")

	// hdrl setup...
	unsigned long hdrl_bytes = 4; // ...4 data bytes for the list 4cc
	unsigned long hdrl_position = _f.Getpos();
	// leave space for the hdrl header...
	if (_f.Seek(12,ivaIO::CUR) == ivaIO::sSeekOffsetError) FAIL(me, "hdrl", "seek failed")

	// adjust all avi header values
	if (_streams.size() > 0)
		{
		iva_WriteStreamBase	*s	= _video_streams.size() > 0 ? 
							(iva_WriteStreamBase *)_video_streams[0] : 
							(iva_WriteStreamBase *)_streams[0];
		const ivaStreamHeader *h	= s->header();
		if (h->dwRate && h->dwScale)
			{
			_header->dwMicroSecPerFrame = (int)((((float)1000000) * h->dwScale) / h->dwRate);
			_header->dwSuggestedBufferSize = s->max_frame() * h->dwRate / h->dwScale;
			_header->dwMaxBytesPerSec	= s->max_frame() * h->dwRate / h->dwScale;
			}
		}
	else
		{
		_header->dwMicroSecPerFrame = 1000000;
		_header->dwWidth			= 0;
		_header->dwHeight			= 0;
		_header->dwSuggestedBufferSize = 1000;
		_header->dwMaxBytesPerSec = 1000;
		}

	if (_video_streams.size() > 0)
		{
		_header->dwWidth			= _video_streams[0]->stream_format()->bmiHeader.biWidth;
		_header->dwHeight			= _video_streams[0]->stream_format()->bmiHeader.biHeight;
		}
	//_header->dwPaddingGranularity;
	//_header->dwFlags;
	_header->dwTotalFrames		= (_streams.size() > 0 ? 
									_streams[0]->frames_written(): 0);
	//_header->dwInitialFrames	= ;
	_header->dwStreams			= _streams.size();


	//////


	// avi header data...
	unsigned long avih_bytes = ivaHeader::size();
	ffirChunkHeader avih_header(ffirFCC::avih(), avih_bytes);
	if (!avih_header.write(_f)) FAIL(me, "avih header", "write failed") 

	if (!_header->write(_f)) FAIL(me, "avih", "write failed") 
	hdrl_bytes += 8 + ivaHeader::size();

	// stream headers...
	i = _streams.begin();
	while (i != _streams.end()) {
		unsigned long strl_bytes;
		if (!(*i)->write_header(_f, strl_bytes)) FAIL(me, "strl", "write failed")
		hdrl_bytes += strl_bytes;
		++i;
	}

	// padding...
	unsigned long header_bytes = hdrl_bytes  // ...including list 4cc
		+ 12 // ...bytes for the riff header & list 4cc
		+ 8; // ...bytes for the hdrl header
	unsigned long padding_required = _header_padding - header_bytes - 12;
		// ...last 12 bytes is for the movi chunk header and list 4cc
	long junk_size = padding_required - 8; // ...8 bytes for junk chunk header
	if (junk_size < 0) {
		// disaster...not enough room was reserved for the header!
		FAIL(me, "file header", "insufficient space reserved")
	}
	ffirChunkHeader jh(ffirFCC::JUNK(), junk_size);
	if (!jh.write(_f)) FAIL(me, "junk header", "write failed");
	if (_f.Seek(junk_size,ivaIO::CUR) == ivaIO::sSeekOffsetError) 
		FAIL(me, "alignment", "seek failed")

	// movi header...
	ffirListChunkHeader movi_header(_movi_bytes, ffirFCC::movi());
	if (!movi_header.write(_f)) FAIL(me, "movi header", "write failed")

	// hdrl...
	if (_f.Seek(hdrl_position,ivaIO::SET) == ivaIO::sSeekOffsetError) FAIL(me, "hdrl", "seek failed")
	ffirListChunkHeader hdrl_header(hdrl_bytes, ffirFCC::hdrl());
	if (!hdrl_header.write(_f)) FAIL(me, "hdrl header", "write failed")

	if (!_f.Close()) FAIL(me, _fname, "close failed")

	return 1;
}

int
iva_Writer::write_word_aligned(const uchar *data, size_t &size)
{
	if (!size) return 1;
	if (_f.Write((void *)data, size) != size) return 0;
	// word alignment...
	if (size % 2) {
		if (_f.Seek(1,ivaIO::CUR) == ivaIO::sSeekOffsetError) return 0;
		size += 1;
	}
	return 1;
}

int
iva_WriteStreamBase::write_header(ivaIO &f, unsigned long &bytes_written)
{
	const char *me = "iva_WriteStream::write_header";
	unsigned long strl_data_bytes = 4 // ...list 4cc code
		+ ivaStreamHeader::size() + 8
		+ stream_format_chunk_data_size() + 8
		+ (_driver_data ? _driver_data->length() + 8 : 0);

	ffirListChunkHeader strl_header(strl_data_bytes, ffirFCC::strl());
	if (!strl_header.write(f)) FAIL(me, "strl header", "write failed")

	ffirChunkHeader strh_header(ffirFCC::strh(), ivaStreamHeader::size());
	if (!strh_header.write(f)) FAIL(me, "strh header", "write failed")
	if (!_header->write(f)) FAIL(me, "strh", "write failed") 

	if (!write_stream_format(f)) FAIL(me, "strf", "write failed")

	if (_driver_data) {
		size_t data_size = _driver_data->length();
		ffirChunkHeader strd_header(ffirFCC::strf(), data_size);
		if (!strd_header.write(f)) FAIL(me, "strd header", "write failed")
		if (!_writer->write_word_aligned(_driver_data->buffer(), data_size))
			 FAIL(me, "other header", "write failed")
	}

	bytes_written = strl_data_bytes + 8; // ...data + chunk header
	return 1;
}

iva_Writer::~iva_Writer()
{
	if (_f.IsOpen())
		close();
	delete [] _fname;
	iva_WriteStreamBase_t::iterator i = _streams.begin();
	while (i != _streams.end()) {
		delete (*i);
		++i;
	}
	delete _header;
}
