/*  Copyright (c) 1994-2002 by On2 Technologies, The Duck Corporation Inc.

    This file is part of VpVision.

    VpVision is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License ver. 2,
    as published by the Free Software Foundation (www.fsf.org).

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VpVision; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef iva_Writer_included
#define iva_Writer_included

// $Id: iva_Writer.h,v 1.1 2001/12/12 18:47:52 nradisch Exp $
// Reference implementation of avi file writer service.
//
// Alan S. Rojer
// The Duck Corp.
// 1 Sept. 1997

#include "iva.h"
#include "ffirFCC.h"
#include "iva_Parser.h"
#include <vector>
using namespace std;

class ivaIO;

typedef pmsio_AVIINDEXENTRY iva_IndexEntry;

class iva_Writer;

class iva_WriteStreamBase : public ivaWriteStream {
protected:
	iva_Writer *_writer; // ...reference
	ivaStreamHeader *_header; // ...owned
	ivaDriverData *_driver_data; // ...owned
	int _stream_number;
	size_t _frames_written;
	size_t _bytes_written;
	size_t _max_frame;

public:
	// factory...
	virtual ~iva_WriteStreamBase();
	iva_WriteStreamBase(iva_Writer *, int stream_number, 
		const ivaStreamHeader *, const ivaDriverData*);
public:
	// access methods...
	// header() should not be const because we may have to adjust
	// the header information just before writing it out (ie recalculating
	// the framerate based on how much data was really written)
	virtual	void	header_fixup(void) = 0;
	ivaStreamHeader *header() const {return _header;}
	const ivaDriverData *driver_data() const {return _driver_data;}
public:
	// ivaWriteStream method...
	int write(const ivaFrame &, int is_key, ffirTCC twocc);
	virtual ivaFrame *write_getframe(size_t siz);
	virtual	int		write_returnframe(ivaFrame *f,int is_key, ffirTCC twocc);
public:
	// for use by the containing iva_Writer...
	int write_header(ivaIO &f, unsigned long &bytes_written);
	size_t frames_written() const {return _frames_written;}
	size_t bytes_written() const { return _bytes_written; }
	size_t max_frame() const { return _max_frame; }
public:
	// for the derived (template) classes...
	virtual int write_stream_format(ivaIO &f) = 0;
	virtual size_t stream_format_chunk_data_size() const = 0;
};

template <class StreamFormat>
class iva_WriteStream : public iva_WriteStreamBase {
protected:
	StreamFormat *_stream_format;
public:
	~iva_WriteStream() {delete _stream_format;}
	iva_WriteStream(iva_Writer *, int stream_number, 
		const ivaStreamHeader *,
		const StreamFormat *,
		const ivaDriverData *);
	StreamFormat *stream_format() const {return _stream_format;}
	virtual	void	header_fixup(void);
	int write_stream_format(ivaIO &f);
		// ...template parameter must support 'size_t length() const'
		// and 'int write(FILE*)' methods
	size_t stream_format_chunk_data_size() const {
		size_t native_length = _stream_format->length();
		return (native_length % 2 ? native_length + 1 : native_length);
	}
};

typedef iva_WriteStream<ivaVideoHeader> iva_VideoWriteStream;
typedef iva_WriteStream<ivaAudioHeader> iva_AudioWriteStream;
typedef iva_WriteStream<ivaOtherHeader> iva_OtherWriteStream;

class iva_Writer : public ivaWriter {
protected:
	// member variables...
	ivaIO &_f;
	char *_fname;
	ivaHeader *_header;
#ifdef NEEDS_STL_ALLOCATOR
	typedef vector<iva_AudioWriteStream*,allocator<iva_AudioWriteStream*> > iva_AudioWriteStream_t;
	typedef vector<iva_VideoWriteStream*,allocator<iva_VideoWriteStream*> > iva_VideoWriteStream_t;
	typedef vector<iva_OtherWriteStream*,allocator<iva_OtherWriteStream*> > iva_OtherWriteStream_t;
	typedef vector<iva_WriteStreamBase*,allocator<iva_WriteStreamBase*> > iva_WriteStreamBase_t;
#else
	typedef vector<iva_AudioWriteStream*> iva_AudioWriteStream_t;
	typedef vector<iva_VideoWriteStream*> iva_VideoWriteStream_t;
	typedef vector<iva_OtherWriteStream*> iva_OtherWriteStream_t;
	typedef vector<iva_WriteStreamBase*> iva_WriteStreamBase_t;
#endif

	iva_AudioWriteStream_t _audio_streams;
	iva_VideoWriteStream_t _video_streams;
	iva_OtherWriteStream_t _other_streams;
	iva_WriteStreamBase_t _streams;
	unsigned long _movi_bytes;
	unsigned long _idx1_bytes;
#ifdef NEEDS_STL_ALLOCATOR
	typedef vector<iva_IndexEntry, allocator<iva_IndexEntry> > iva_IndexEntry_t;
#else
	typedef vector<iva_IndexEntry> iva_IndexEntry_t;
#endif
	iva_IndexEntry_t _index;
	size_t _header_padding;
	size_t _chunk_padding;

	void	recalc_framerate(void);
public:
	// factory methods...
	iva_Writer(ivaIO &io,const ivaHeader &header, 
		size_t header_padding, size_t chunk_padding) 
	: _f(io), _header(new ivaHeader(header)),
		_fname(0), _movi_bytes(0), _idx1_bytes(0),
		 _header_padding(header_padding),
		_chunk_padding(chunk_padding){}
	~iva_Writer();
public:
	// ivaFile methods...
	bool open(const char *fname);
	int is_open() const; 
	int close(int re_fr = 0);
	const ivaHeader *header() const {return _header;}
public:
	// ivaWriter methods...
	ivaWriteStream* create_stream(
		const ivaStreamHeader *,
		const ivaVideoHeader *,
		const ivaDriverData *p_dd = 0);
	ivaWriteStream* create_stream(
		const ivaStreamHeader *,
		const ivaAudioHeader *,
		const ivaDriverData *p_dd = 0);
	ivaWriteStream* create_stream(
		const ivaStreamHeader *,
		const ivaOtherHeader *,
		const ivaDriverData *p_dd = 0);
#if(0)
	const ivaWriteStream *audio_stream(size_t i) const {
		if (i < _audio_streams.Pos()) return _audio_streams[i];
		return 0;
	}
	const ivaWriteStream *video_stream(size_t i) const {
		if (i < _video_streams.Pos()) return _video_streams[i];
		return 0;
	}
	const ivaWriteStream *other_stream(size_t i) const {
		if (i < _other_streams.Pos()) return _other_streams[i];
		return 0;
	}
#endif
public:
	// for use by the other guys in here...
	virtual int put_frame(ffirFCC, const ivaFrame &, int is_key);
	virtual	ivaFrame *get_frame(size_t siz);
	virtual int		return_frame(ffirFCC, ivaFrame *f,int is_key);

	int write_word_aligned(const uchar *, size_t &);
		// ...last argument is modified to indicate bytes actually written
};

#endif // iva_Writer_included
